# coding: utf-8
"""

"""

import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms import StringField, SelectField
from wtforms.validators import Length
import typing

from . import frontend
from ..logic import errors
from ..logic.locations import Location, create_location, get_location, get_locations_tree, update_location, get_object_location_assignment, confirm_object_responsibility
from ..logic.security_tokens import verify_token
from ..logic.notifications import mark_notification_for_being_assigned_as_responsible_user_as_read
from .utils import check_current_user_is_not_readonly


class LocationForm(FlaskForm):
    name = StringField(validators=[Length(min=1, max=100)])
    description = StringField()
    parent_location = SelectField()


@frontend.route('/locations/')
@flask_login.login_required
def locations():
    locations_map, locations_tree = get_locations_tree()
    return flask.render_template(
        'locations/locations.html',
        locations_map=locations_map, locations_tree=locations_tree,
        sort_location_ids_by_name=_sort_location_ids_by_name
    )


@frontend.route('/locations/<int:location_id>', methods=['GET', 'POST'])
@flask_login.login_required
def location(location_id):
    try:
        location = get_location(location_id)
    except errors.LocationDoesNotExistError:
        return flask.abort(404)
    mode = flask.request.args.get('mode', None)
    if mode == 'edit':
        if flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_LOCATIONS'] and not flask_login.current_user.is_admin:
            flask.flash('Only administrators can edit locations.', 'error')
            return flask.abort(403)
        check_current_user_is_not_readonly()
        return _show_location_form(location, None)
    locations_map, locations_tree = get_locations_tree()
    ancestors = []
    parent_location = location
    while parent_location.parent_location_id is not None:
        parent_location = get_location(parent_location.parent_location_id)
        ancestors.insert(0, (parent_location.id, parent_location.name))
    for ancestor_id, ancestor_name in ancestors:
        locations_tree = locations_tree[ancestor_id]
    locations_tree = locations_tree[location_id]
    return flask.render_template(
        'locations/location.html',
        locations_map=locations_map, locations_tree=locations_tree,
        location=location, ancestors=ancestors,
        sort_location_ids_by_name=_sort_location_ids_by_name
    )


@frontend.route('/locations/new/', methods=['GET', 'POST'])
@flask_login.login_required
def new_location():
    if flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_LOCATIONS'] and not flask_login.current_user.is_admin:
        flask.flash('Only administrators can create locations.', 'error')
        return flask.abort(403)
    check_current_user_is_not_readonly()
    parent_location = None
    parent_location_id = flask.request.args.get('parent_location_id', None)
    if parent_location_id is not None:
        try:
            parent_location_id = int(parent_location_id)
        except ValueError:
            parent_location_id = None
    if parent_location_id:
        try:
            parent_location = get_location(parent_location_id)
        except errors.LocationDoesNotExistError:
            flask.flash('The requested parent location does not exist.', 'error')
    return _show_location_form(None, parent_location)


@frontend.route('/locations/confirm_responsibility')
@flask_login.login_required
def accept_responsibility_for_object():
    token = flask.request.args.get('t', None)
    if token is None:
        flask.flash('The confirmation token is missing.', 'error')
        return flask.redirect(flask.url_for('.index'))
    object_location_assignment_id = verify_token(token, salt='confirm_responsibility', secret_key=flask.current_app.config['SECRET_KEY'], expiration=None)
    if object_location_assignment_id is None:
        flask.flash('The confirmation token is invalid.', 'error')
        return flask.redirect(flask.url_for('.index'))
    try:
        object_location_assignment = get_object_location_assignment(object_location_assignment_id)
    except errors.ObjectLocationAssignmentDoesNotExistError:
        flask.flash('This responsibility assignment does not exist.', 'error')
        return flask.redirect(flask.url_for('.index'))
    if object_location_assignment.responsible_user_id != flask_login.current_user.id:
        flask.flash('This responsibility assignment belongs to another user.', 'error')
        return flask.redirect(flask.url_for('.index'))
    if object_location_assignment.confirmed:
        flask.flash('This responsibility assignment has already been confirmed.', 'success')
    else:
        confirm_object_responsibility(object_location_assignment_id)
        flask.flash('You have successfully confirmed this responsibility assignment.', 'success')
        mark_notification_for_being_assigned_as_responsible_user_as_read(
            user_id=flask_login.current_user.id,
            object_location_assignment_id=object_location_assignment_id
        )
    return flask.redirect(flask.url_for('.object', object_id=object_location_assignment.object_id))


def _sort_location_ids_by_name(location_ids: typing.Iterable[int], location_map: typing.Dict[int, Location]) -> typing.List[int]:
    location_ids = list(location_ids)
    location_ids.sort(key=lambda location_id: location_map[location_id].name)
    return location_ids


def _show_location_form(location: typing.Optional[Location], parent_location: typing.Optional[Location]):
    if location is not None:
        submit_text = "Save"
    elif parent_location is not None:
        submit_text = "Create"
    else:
        submit_text = "Create"
    locations_map, locations_tree = get_locations_tree()
    invalid_location_ids = []
    if location is not None:
        invalid_location_ids.append(location.id)
        ancestor_ids = []
        _parent_location = location
        while _parent_location.parent_location_id is not None:
            _parent_location = get_location(_parent_location.parent_location_id)
            ancestor_ids.insert(0, _parent_location.id)
        locations_subtree = locations_tree
        for ancestor_id in ancestor_ids:
            locations_subtree = locations_subtree[ancestor_id]
        locations_subtree = locations_subtree[location.id]
        unhandled_descendent_ids_and_subtrees = [(descendent_id, locations_subtree) for descendent_id in locations_subtree]
        while unhandled_descendent_ids_and_subtrees:
            descendent_id, locations_subtree = unhandled_descendent_ids_and_subtrees.pop(0)
            invalid_location_ids.append(descendent_id)
            locations_subtree = locations_subtree[descendent_id]
            for descendent_id in locations_subtree:
                unhandled_descendent_ids_and_subtrees.append((descendent_id, locations_subtree))

    location_form = LocationForm()
    location_form.parent_location.choices = [('-1', '-')] + [
        (str(location_id), locations_map[location_id].name)
        for location_id in locations_map
        if location_id not in invalid_location_ids
    ]
    if location_form.parent_location.data is None or location_form.parent_location.data == str(None):
        if location is not None and location.parent_location_id:
            location_form.parent_location.data = str(location.parent_location_id)
        elif parent_location is not None:
            location_form.parent_location.data = str(parent_location.id)
        else:
            location_form.parent_location.data = str(-1)

    form_is_valid = False
    if location_form.validate_on_submit():
        form_is_valid = True

    if location is not None:
        if location_form.name.data is None:
            location_form.name.data = location.name
        if location_form.description.data is None:
            location_form.description.data = location.description

    if form_is_valid:
        name = location_form.name.data
        description = location_form.description.data
        parent_location_id = location_form.parent_location.data
        try:
            parent_location_id = int(parent_location_id)
        except ValueError:
            parent_location_id = None
        if parent_location_id < 0:
            parent_location_id = None
        if location is None:
            location = create_location(name, description, parent_location_id, flask_login.current_user.id)
            flask.flash('The location was created successfully.', 'success')

        else:
            update_location(location.id, name, description, parent_location_id, flask_login.current_user.id)
            flask.flash('The location was updated successfully.', 'success')
        return flask.redirect(flask.url_for('.location', location_id=location.id))
    return flask.render_template(
        'locations/location_form.html',
        location_form=location_form,
        submit_text=submit_text
    )



# coding: utf-8
"""

"""


class ObjectDoesNotExistError(Exception):
    pass


class FileDoesNotExistError(Exception):
    pass


class ObjectVersionDoesNotExistError(Exception):
    pass


class FileNameTooLongError(Exception):
    pass


class TooManyFilesForObjectError(Exception):
    pass


class InvalidFileStorageError(Exception):
    pass


class GroupDoesNotExistError(Exception):
    pass


class GroupAlreadyExistsError(Exception):
    pass


class UserDoesNotExistError(Exception):
    pass


class UserNotMemberOfGroupError(Exception):
    pass


class UserAlreadyMemberOfGroupError(Exception):
    pass


class InvalidGroupNameError(Exception):
    pass


class ActionDoesNotExistError(Exception):
    pass


class InstrumentDoesNotExistError(Exception):
    pass


class UserAlreadyResponsibleForInstrumentError(Exception):
    pass


class UserNotResponsibleForInstrumentError(Exception):
    pass


class UndefinedUnitError(Exception):
    pass


class ProjectAlreadyExistsError(Exception):
    pass


class InvalidProjectNameError(Exception):
    pass


class ProjectDoesNotExistError(Exception):
    pass


class UserNotMemberOfProjectError(Exception):
    pass


class UserAlreadyMemberOfProjectError(Exception):
    pass


class GroupAlreadyMemberOfProjectError(Exception):
    pass


class GroupNotMemberOfProjectError(Exception):
    pass


class NoMemberWithGrantPermissionsForProjectError(Exception):
    pass


class InvalidSubprojectRelationshipError(Exception):
    pass


class SubprojectRelationshipDoesNotExistError(Exception):
    pass


class SchemaError(Exception):
    def __init__(self, message, path):
        if path:
            message += ' (at ' + ' -> '.join(path) + ')'
        super(SchemaError, self).__init__(message)
        self.path = path
        self.message = message
        self.paths = [path]


class ValidationError(SchemaError):
    def __init__(self, message, path):
        super(ValidationError, self).__init__(message, path)


class ValidationMultiError(ValidationError):
    def __init__(self, errors):
        message = '\n'.join(error.message for error in errors)
        super(ValidationMultiError, self).__init__(message, [])
        self.paths = [error.path for error in errors]


class LocationDoesNotExistError(Exception):
    pass


class CyclicLocationError(Exception):
    pass


class ObjectLocationAssignmentDoesNotExistError(Exception):
    pass


class NotificationDoesNotExistError(Exception):
    pass


class OnlyOneAuthenticationMethod(Exception):
    pass


class AuthenticationMethodWrong(Exception):
    pass


class AuthenticationMethodAlreadyExists(Exception):
    pass


class NoEmailInLDAPAccountError(Exception):
    pass


class LDAPAccountAlreadyExistError(Exception):
    pass


class LDAPAccountOrPasswordWrongError(Exception):
    pass


class PublicationDoesNotExistError(Exception):
    pass


class InvalidDOIError(Exception):
    pass


class LDAPNotConfiguredError(Exception):
    pass


class UserIsReadonlyError(Exception):
    pass


class InstrumentLogEntryDoesNotExistError(Exception):
    pass


class InstrumentLogFileAttachmentDoesNotExistError(Exception):
    pass


class InstrumentLogObjectAttachmentDoesNotExistError(Exception):
    pass


class InstrumentLogCategoryDoesNotExistError(Exception):
    pass


class GroupInvitationDoesNotExistError(Exception):
    pass


class ProjectInvitationDoesNotExistError(Exception):
    pass


class UserInvitationDoesNotExistError(Exception):
    pass


class ActionTypeDoesNotExistError(Exception):
    pass


class CommentDoesNotExistError(Exception):
    pass



# coding: utf-8
"""
Model for instrument log entries

This should not to be confused with internal logging like the object log.
"""

import enum
import datetime
import typing

from .. import db
from .instruments import Instrument
from .objects import Objects


instrument_log_entry_category_association_table = db.Table(
    'instrument_log_entry_version_category_associations',
    db.metadata,
    db.Column('log_entry_id', db.Integer),
    db.Column('log_entry_version_id', db.Integer),
    db.Column('category_id', db.Integer, db.ForeignKey('instrument_log_categories.id', ondelete="CASCADE")),
    db.ForeignKeyConstraint(
        ['log_entry_id', 'log_entry_version_id'],
        ['instrument_log_entry_versions.log_entry_id', 'instrument_log_entry_versions.version_id']
    )
)


@enum.unique
class InstrumentLogCategoryTheme(enum.Enum):
    GRAY = 0
    BLUE = 1
    GREEN = 2
    YELLOW = 3
    RED = 4


class InstrumentLogCategory(db.Model):
    __tablename__ = 'instrument_log_categories'

    id = db.Column(db.Integer, primary_key=True)
    instrument_id = db.Column(db.Integer, db.ForeignKey(Instrument.id), nullable=False)
    title = db.Column(db.String, nullable=False)
    theme = db.Column(db.Enum(InstrumentLogCategoryTheme), nullable=False)

    def __init__(self, instrument_id: int, title: str, theme: InstrumentLogCategoryTheme):
        self.instrument_id = instrument_id
        self.title = title
        self.theme = theme

    def __repr__(self):
        return f'<{type(self).__name__}(id={self.id}, instrument_id={self.instrument_id}, title="{self.title}", theme={self.theme.name.lower()})>'


class InstrumentLogEntry(db.Model):
    __tablename__ = 'instrument_log_entries'

    id = db.Column(db.Integer, primary_key=True)
    instrument_id = db.Column(db.Integer, db.ForeignKey(Instrument.id), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    author = db.relationship('User')

    def __init__(
            self,
            instrument_id: int,
            user_id: int
    ):
        self.instrument_id = instrument_id
        self.user_id = user_id

    def __repr__(self):
        return '<{0}(id={1.id}, instrument_id={1.instrument_id}, user_id={1.user_id})>'.format(type(self).__name__, self)


class InstrumentLogEntryVersion(db.Model):
    __tablename__ = 'instrument_log_entry_versions'

    log_entry_id = db.Column(db.Integer, db.ForeignKey(InstrumentLogEntry.id), primary_key=True)
    version_id = db.Column(db.Integer, primary_key=True)
    content = db.Column(db.Text, nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)
    categories = db.relationship('InstrumentLogCategory', secondary=instrument_log_entry_category_association_table)
    log_entry = db.relationship(InstrumentLogEntry, backref='versions')

    def __init__(
            self,
            log_entry_id: int,
            version_id: int,
            content: str,
            utc_datetime: typing.Optional[datetime.datetime] = None
    ):
        self.log_entry_id = log_entry_id
        self.version_id = version_id
        self.content = content
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        self.utc_datetime = utc_datetime

    def __repr__(self):
        return '<{0}(log_entry_id={1.log_entry_id}, version_id={1.version_id},  utc_datetime={1.utc_datetime}, content="{1.content}")>'.format(type(self).__name__, self)


class InstrumentLogFileAttachment(db.Model):
    __tablename__ = 'instrument_log_file_attachments'

    id = db.Column(db.Integer, primary_key=True)
    log_entry_id = db.Column(db.Integer, db.ForeignKey(InstrumentLogEntry.id), nullable=False)
    file_name = db.Column(db.String, nullable=False)
    content = db.Column(db.LargeBinary, nullable=False)
    is_hidden = db.Column(db.Boolean, default=False, nullable=False)

    def __init__(self, log_entry_id: int, file_name: str, content: bytes):
        self.log_entry_id = log_entry_id
        self.file_name = file_name
        self.content = content

    def __repr__(self):
        return '<{0}(id={1.id}, log_entry_id={1.log_entry_id}, file_name="{1.file_name}")>'.format(type(self).__name__, self)


class InstrumentLogObjectAttachment(db.Model):
    __tablename__ = 'instrument_log_object_attachments'

    id = db.Column(db.Integer, primary_key=True)
    log_entry_id = db.Column(db.Integer, db.ForeignKey(InstrumentLogEntry.id), nullable=False)
    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    is_hidden = db.Column(db.Boolean, default=False, nullable=False)

    def __init__(self, log_entry_id: int, object_id: int):
        self.log_entry_id = log_entry_id
        self.object_id = object_id

    def __repr__(self):
        return '<{0}(id={1.id}, log_entry_id={1.log_entry_id}, object_id={1.object_id})>'.format(type(self).__name__, self)



from wtforms.validators import ValidationError

import flask
import flask_login

from ..models import Permissions
from ..logic.object_permissions import get_user_object_permissions


def ObjectIdValidator(required_perm: Permissions, allow_self: bool = False):
    def validate(form, field):
        user_id = flask_login.current_user.id
        object_id = field.data

        try:
            object_id = int(object_id)
        except ValueError:
            raise ValidationError("Object_id is not an int.")

        if "object_id" in flask.request.view_args and not allow_self:
            if object_id == flask.request.view_args["object_id"]:
                raise ValidationError("You can not select the same object.")

        perm = get_user_object_permissions(object_id=object_id, user_id=user_id)
        if required_perm > perm:
            raise ValidationError("No such object or no access to object")

    return validate


def MultipleObjectIdValidator(required_perm: Permissions, allow_self: bool = False):
    def validate(form, field):
        user_id = flask_login.current_user.id
        object_ids = field.data
        for object_id in object_ids:
            try:
                object_id = int(object_id)
            except ValueError:
                raise ValidationError("Object_id is not an int.")

            if "object_id" in flask.request.view_args and not allow_self:
                if object_id == flask.request.view_args["object_id"]:
                    raise ValidationError("You can not select the same object.")

            perm = get_user_object_permissions(object_id=object_id, user_id=user_id)
            if required_perm > perm:
                raise ValidationError("No such object or no access to object")

    return validate



# coding: utf-8
"""

"""

import base64
import functools
import json
import os
import typing

import flask
import flask_login
import sqlalchemy

from . import logic, db
from .models import Permissions, migrations

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def ansi_color(text: str, color: int):
    """
    Add ANSI color codes to text.

    :param text: the text without ANSI colors
    :param color: the desired color
    :return: the text with the color added
    """
    return f'\033[{color}m{text}\033[0m'


def object_permissions_required(
        required_object_permissions: Permissions,
        auth_extension: typing.Any = flask_login,
        user_id_callable: typing.Callable[[], int] = lambda: flask_login.current_user.id,
        on_unauthorized: typing.Callable[[int], None] = lambda object_id: flask.abort(403)
):
    def decorator(func, user_id_callable=user_id_callable, on_unauthorized=on_unauthorized):
        @auth_extension.login_required
        @functools.wraps(func)
        def wrapper(*args, user_id_callable=user_id_callable, on_unauthorized=on_unauthorized, **kwargs):
            assert 'object_id' in kwargs
            object_id = kwargs['object_id']
            version_id = kwargs.get('version_id')
            try:
                logic.objects.get_object(object_id, version_id)
            except logic.errors.ObjectDoesNotExistError:
                return flask.abort(404)
            except logic.errors.ObjectVersionDoesNotExistError:
                return flask.abort(404)
            if not (logic.object_permissions.object_is_public(object_id) and required_object_permissions in Permissions.READ):
                user_id = user_id_callable()
                user = logic.users.get_user(user_id)
                if user.is_readonly and required_object_permissions not in Permissions.READ:
                    return on_unauthorized(object_id)
                user_object_permissions = logic.object_permissions.get_user_object_permissions(object_id=object_id, user_id=user_id)
                if required_object_permissions not in user_object_permissions:
                    return on_unauthorized(object_id)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def load_environment_configuration(env_prefix):
    """
    Loads configuration data from environment variables with a given prefix.
    If the prefixed environment variable B64_JSON_ENV exists, its content
    will be treated as an Base64 encoded JSON object and it's attributes
    starting with the prefix will be added to the environment.

    :return: a dict containing the configuration values
    """
    b64_json_env = os.environ.get(env_prefix + 'B64_JSON_ENV', None)
    if b64_json_env:
        for key, value in json.loads(base64.b64decode(b64_json_env.encode('utf-8')).decode('utf-8')).items():
            if key.startswith(env_prefix):
                os.environ[key] = value
    config = {
        key[len(env_prefix):]: value
        for key, value in os.environ.items()
        if key.startswith(env_prefix) and key != env_prefix + 'B64_JSON_ENV'
    }
    return config


def generate_secret_key(num_bits):
    """
    Generates a secure, random key for the application.

    :param num_bits: number of bits of random data in the secret key
    :return: the base64 encoded secret key
    """
    num_bytes = num_bits // 8
    binary_key = os.urandom(num_bytes)
    base64_key = base64.b64encode(binary_key).decode('ascii')
    return base64_key


def empty_database(engine, recreate=False):
    metadata = sqlalchemy.MetaData(bind=engine)
    # delete views, as SQLAlchemy cannot reflect them
    engine.execute("DROP VIEW IF EXISTS user_object_permissions_by_all")
    # delete tables, etc
    metadata.reflect()
    metadata.drop_all()
    if recreate:
        # recreate the tables
        db.metadata.create_all(bind=engine)
        # run migrations
        migrations.run(db)



# coding: utf-8
"""

"""

import flask
import flask_login

from .. import frontend
from ...logic.users import get_users


@frontend.route('/users/')
@flask_login.login_required
def users():
    return flask.render_template(
        'users.html',
        users=sorted(get_users(exclude_hidden=not flask_login.current_user.is_admin), key=lambda u: u.id)
    )



# coding: utf-8
"""

"""

import enum
import flask_login
from .. import db


@enum.unique
class UserType(enum.Enum):
    PERSON = 1
    OTHER = 2


class User(db.Model, flask_login.UserMixin):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, nullable=False)
    email = db.Column(db.String, nullable=False)
    type = db.Column(db.Enum(UserType), nullable=False)
    is_admin = db.Column(db.Boolean, default=False, nullable=False)
    is_readonly = db.Column(db.Boolean, default=False, nullable=False)
    is_hidden = db.Column(db.Boolean, default=False, nullable=False)
    orcid = db.Column(db.String, nullable=True)
    affiliation = db.Column(db.String, nullable=True)
    is_active = db.Column(db.Boolean, default=True, nullable=False)

    def __init__(self, name, email, type):
        self.name = name
        self.email = email
        self.type = type

    def __eq__(self, other):
        try:
            return (
                self.id == other.id and
                self.name == other.name and
                self.email == other.email and
                self.type == other.type and
                self.is_admin == other.is_admin and
                self.is_readonly == other.is_readonly and
                self.is_hidden == other.is_hidden and
                self.orcid == other.orcid
            )
        except AttributeError:
            return False

    def get_id(self):
        return self.id

    def __repr__(self):
        return '<{0}(id={1.id}, name={1.name})>'.format(type(self).__name__, self)


class UserInvitation(db.Model):
    __tablename__ = 'user_invitations'

    id = db.Column(db.Integer, primary_key=True)
    inviter_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)
    accepted = db.Column(db.Boolean, nullable=False, default=False)



# coding: utf-8
"""
Create the view user_object_permissions_by_all to speed up object requests
with minimum permissions.
"""

import os

MIGRATION_INDEX = 7
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Perform migration
    db.engine.execute("""
    CREATE OR REPLACE VIEW user_object_permissions_by_all
    AS
        SELECT
        NULL AS user_id,
        object_id AS object_id,
        1 AS permissions_int
        FROM public_objects
    UNION
        SELECT
        association.user_id AS user_id,
        objects.object_id AS object_id,
        3 AS permissions_int
        FROM association AS association
        JOIN actions AS actions ON association.instrument_id = actions.instrument_id
        JOIN objects_current AS objects ON objects.action_id = actions.id
    UNION
        SELECT
        user_object_permissions.user_id AS user_id,
        user_object_permissions.object_id AS object_id,
        ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> user_object_permissions.permissions::text)::int AS permissions_int
        FROM user_object_permissions
    UNION
        SELECT
        user_group_memberships.user_id AS user_id,
        group_object_permissions.object_id AS object_id,
        ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> group_object_permissions.permissions::text)::int AS permissions_int
        FROM user_group_memberships AS user_group_memberships
        JOIN group_object_permissions AS group_object_permissions ON user_group_memberships.group_id = group_object_permissions.group_id
    UNION
        SELECT
        user_project_permissions.user_id AS user_id,
        project_object_permissions.object_id AS object_id,
        LEAST(
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> project_object_permissions.permissions::text)::int,
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> user_project_permissions.permissions::text)::int
        ) AS permissions_int
        FROM user_project_permissions AS user_project_permissions
        JOIN project_object_permissions AS project_object_permissions ON user_project_permissions.project_id = project_object_permissions.project_id
    UNION
        SELECT
        user_group_memberships.user_id AS user_id,
        project_object_permissions.object_id AS object_id,
        LEAST(
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> project_object_permissions.permissions::text)::int,
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> group_project_permissions.permissions::text)::int
        ) AS permissions_int
        FROM user_group_memberships AS user_group_memberships
        JOIN group_project_permissions ON group_project_permissions.group_id = user_group_memberships.group_id
        JOIN project_object_permissions AS project_object_permissions ON group_project_permissions.project_id = project_object_permissions.project_id
    """)
    return True



# coding: utf-8
"""
Add migration_indices table to database.
"""

import os

MIGRATION_INDEX = 0
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    migration_index_exists = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'migration_index'
    """).first() is not None
    if migration_index_exists:
        return False

    # Perform migration
    db.session.execute("""
        CREATE TABLE migration_index (
          migration_index INTEGER
        )
    """)

    # Insert migration index 0
    db.session.execute(
        """
        INSERT INTO migration_index
        (migration_index) VALUES (0)
        """
    )
    return True



# coding: utf-8
"""

"""

import flask
import flask_login

from . import frontend
from ..logic.tags import get_tags


@frontend.route('/tags/')
@flask_login.login_required
def tags():
    return flask.render_template(
        'tags.html',
        tags=list(sorted(get_tags(), key=lambda tag: (tag.name, tag.uses, tag.id)))
    )



# coding: utf-8
"""

"""

import collections
import datetime
import typing

import flask

from .. import db
from . import errors
from .. models import users, User, UserType


class UserInvitation(collections.namedtuple('UserInvitation', ['id', 'inviter_id', 'utc_datetime', 'accepted'])):
    """
    This class provides an immutable wrapper around models.users.UserInvitation.
    """

    def __new__(cls, id: int, inviter_id: int, utc_datetime: datetime.datetime, accepted: bool):
        self = super(UserInvitation, cls).__new__(cls, id, inviter_id, utc_datetime, accepted)
        return self

    @classmethod
    def from_database(cls, user_invitation: users.UserInvitation) -> 'UserInvitation':
        return UserInvitation(
            id=user_invitation.id,
            inviter_id=user_invitation.inviter_id,
            utc_datetime=user_invitation.utc_datetime,
            accepted=user_invitation.accepted
        )

    @property
    def expired(self):
        expiration_datetime = self.utc_datetime + datetime.timedelta(seconds=flask.current_app.config['INVITATION_TIME_LIMIT'])
        return datetime.datetime.utcnow() >= expiration_datetime


def get_user(user_id: int) -> User:
    if user_id is None:
        raise TypeError("user_id must be int")
    user = User.query.get(user_id)
    if user is None:
        raise errors.UserDoesNotExistError()
    return user


def get_users(exclude_hidden: bool = False) -> typing.List[User]:
    """
    Returns all users.

    :param exclude_hidden: whether or not to exclude hidden users
    :return: the list of users
    """
    if exclude_hidden:
        return User.query.filter_by(is_hidden=False).all()
    return User.query.all()


def get_administrators() -> typing.List[User]:
    """
    Returns all current administrators.

    :return: the list of administrators
    """
    return User.query.filter_by(is_admin=True).all()


def get_users_by_name(name: str) -> typing.List[User]:
    """
    Return all users with a given name.

    :param name: the user name to search for
    :return: the list of users with this name
    """
    return User.query.filter_by(name=name).all()


def create_user(name: str, email: str, type: UserType) -> User:
    """
    Create a new user.

    This function cannot create a user as an administrator. To set whether or
    not a user is an administrator, use the set_administrator script or modify
    the User object returned by this function.

    :param name: the user's name
    :param email: the user's email address
    :param type: the user's type
    :return: the newly created user
    """
    user = User(name=name, email=email, type=type)
    db.session.add(user)
    db.session.commit()
    return user


def set_user_readonly(user_id: int, readonly: bool) -> None:
    """
    Set whether a user should be limited to READ permissions.

    :param user_id: the user ID of an existing user
    :param readonly: True, if the user should be read only, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """

    user = get_user(user_id)
    user.is_readonly = readonly
    db.session.add(user)
    db.session.commit()


def set_user_hidden(user_id: int, hidden: bool) -> None:
    """
    Set whether a user should be hidden from user lists.

    :param user_id: the user ID of an existing user
    :param hidden: True, if the user should be hidden, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """

    user = get_user(user_id)
    user.is_hidden = hidden
    db.session.add(user)
    db.session.commit()


def set_user_active(user_id: int, active: bool) -> None:
    """
    Set whether a user should be allowed to sign in.

    :param user_id: the user ID of an existing user
    :param active: True, if the user should be active, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """

    user = get_user(user_id)
    user.is_active = active
    db.session.add(user)
    db.session.commit()


def set_user_administrator(user_id: int, is_admin: bool) -> None:
    """
    Set whether a user is an administrator.

    :param user_id: the user ID of an existing user
    :param is_admin: True, if the user is an administrator, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """

    user = get_user(user_id)
    user.is_admin = is_admin
    db.session.add(user)
    db.session.commit()


def get_user_invitation(invitation_id: int) -> UserInvitation:
    """
    Get an existing invitation.

    :param invitation_id: the ID of an existing invitation
    :return: the invitation
    :raise errors.UserInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = users.UserInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.UserInvitationDoesNotExistError()
    return UserInvitation.from_database(invitation)


def set_user_invitation_accepted(invitation_id: int) -> None:
    """
    Mark an invitation as having been accepted.

    :param invitation_id: the ID of an existing invitation
    :raise errors.UserInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = users.UserInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.UserInvitationDoesNotExistError()
    invitation.accepted = True
    db.session.add(invitation)
    db.session.commit()



# coding: utf-8
"""

"""

import flask
import flask_login

from . import frontend
from .. import logic
from ..logic.object_permissions import Permissions
from ..logic.security_tokens import verify_token
from .projects_forms import CreateProjectForm, EditProjectForm, LeaveProjectForm, InviteUserToProjectForm, InviteGroupToProjectForm, ProjectPermissionsForm, AddSubprojectForm, RemoveSubprojectForm, DeleteProjectForm, RemoveProjectMemberForm, RemoveProjectGroupForm
from .utils import check_current_user_is_not_readonly


@frontend.route('/projects/<int:project_id>', methods=['GET', 'POST'])
@flask_login.login_required
def project(project_id):
    if 'token' in flask.request.args:
        token = flask.request.args.get('token')
        expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
        token_data = verify_token(token, salt='invite_to_project', secret_key=flask.current_app.config['SECRET_KEY'], expiration=expiration_time_limit)
        if token_data is None:
            flask.flash('Invalid project invitation token. Please request a new invitation.', 'error')
            return flask.abort(403)
        if 'invitation_id' in token_data:
            if logic.projects.get_project_invitation(token_data['invitation_id']).accepted:
                flask.flash('This invitation token has already been used. Please request a new invitation.', 'error')
                return flask.abort(403)
        if token_data.get('project_id', None) != project_id:
            return flask.abort(403)
        user_id = token_data.get('user_id', None)
        if user_id != flask_login.current_user.id:
            if user_id is not None:
                try:
                    invited_user = logic.users.get_user(user_id)
                    flask.flash('Please sign in as user "{}" to accept this invitation.'.format(invited_user.name), 'error')
                except logic.errors.UserDoesNotExistError:
                    pass
            return flask.abort(403)
        other_project_ids = token_data.get('other_project_ids', [])
        for notification in logic.notifications.get_notifications(user_id, unread_only=True):
            if notification.type == logic.notifications.NotificationType.INVITED_TO_PROJECT:
                if notification.data['project_id'] == project_id:
                    logic.notifications.mark_notification_as_read(notification.id)
        try:
            logic.projects.add_user_to_project(project_id, user_id, logic.object_permissions.Permissions.READ, other_project_ids=other_project_ids)
        except logic.errors.UserAlreadyMemberOfProjectError:
            flask.flash('You are already a member of this project', 'error')
        except logic.errors.ProjectDoesNotExistError:
            pass
    user_id = flask_login.current_user.id
    try:
        project = logic.projects.get_project(project_id)
    except logic.errors.ProjectDoesNotExistError:
        return flask.abort(404)
    user_permissions = logic.projects.get_user_project_permissions(project_id=project_id, user_id=user_id, include_groups=True)
    if Permissions.READ in user_permissions:
        show_objects_link = True
    else:
        show_objects_link = False
    if Permissions.READ in user_permissions:
        leave_project_form = LeaveProjectForm()
    else:
        leave_project_form = None
    if Permissions.WRITE in user_permissions:
        edit_project_form = EditProjectForm()
        if edit_project_form.name.data is None:
            edit_project_form.name.data = project.name
        if edit_project_form.description.data is None:
            edit_project_form.description.data = project.description
    else:
        edit_project_form = None
    show_edit_form = False
    project_member_user_ids_and_permissions = logic.projects.get_project_member_user_ids_and_permissions(project_id=project_id, include_groups=False)
    project_member_group_ids_and_permissions = logic.projects.get_project_member_group_ids_and_permissions(project_id=project_id)

    project_member_user_ids = list(project_member_user_ids_and_permissions.keys())
    project_member_user_ids.sort(key=lambda user_id: logic.users.get_user(user_id).name.lower())

    project_member_group_ids = list(project_member_group_ids_and_permissions.keys())
    project_member_group_ids.sort(key=lambda group_id: logic.groups.get_group(group_id).name.lower())

    if Permissions.GRANT in user_permissions:
        invitable_user_list = []
        for user in logic.users.get_users(exclude_hidden=True):
            if user.id not in project_member_user_ids_and_permissions:
                invitable_user_list.append(user)
        parent_projects_with_add_permissions = logic.projects.get_ancestor_project_ids(project_id, only_if_child_can_add_users_to_ancestor=True)
    else:
        invitable_user_list = []
        parent_projects_with_add_permissions = []
    if invitable_user_list:
        other_project_ids_data = []
        for parent_project_id in parent_projects_with_add_permissions:
            other_project_ids_data.append({'project_id': parent_project_id})
        invite_user_form = InviteUserToProjectForm(other_project_ids=other_project_ids_data)
    else:
        invite_user_form = None
    if Permissions.GRANT in user_permissions:
        invitable_group_list = []
        for group in logic.groups.get_user_groups(user_id):
            if group.id not in project_member_group_ids_and_permissions:
                invitable_group_list.append(group)
    else:
        invitable_group_list = []
    if invitable_group_list:
        other_project_ids_data = []
        for parent_project_id in parent_projects_with_add_permissions:
            other_project_ids_data.append({'project_id': parent_project_id})
        invite_group_form = InviteGroupToProjectForm(other_project_ids=other_project_ids_data)
    else:
        invite_group_form = None
    child_project_ids = logic.projects.get_child_project_ids(project_id)
    child_project_ids_can_add_to_parent = {child_project_id: logic.projects.can_child_add_users_to_parent_project(parent_project_id=project_id, child_project_id=child_project_id) for child_project_id in child_project_ids}
    parent_project_ids = logic.projects.get_parent_project_ids(project_id)
    add_subproject_form = None
    remove_subproject_form = None
    delete_project_form = None
    remove_project_member_form = None
    remove_project_group_form = None
    addable_projects = []
    addable_project_ids = []
    if Permissions.GRANT in user_permissions:
        for other_project in logic.projects.get_user_projects(flask_login.current_user.id, include_groups=True):
            if other_project.id == project.id:
                continue
            if Permissions.GRANT in logic.projects.get_user_project_permissions(other_project.id, flask_login.current_user.id, include_groups=True):
                addable_projects.append(other_project)
        addable_project_ids = logic.projects.filter_child_project_candidates(project_id, [child_project.id for child_project in addable_projects])
        addable_projects = [logic.projects.get_project(child_project_id) for child_project_id in addable_project_ids]
        if addable_projects:
            add_subproject_form = AddSubprojectForm()
        if child_project_ids:
            remove_subproject_form = RemoveSubprojectForm()
        delete_project_form = DeleteProjectForm()
        remove_project_member_form = RemoveProjectMemberForm()
        remove_project_group_form = RemoveProjectGroupForm()

    project_invitations = None
    show_invitation_log = flask_login.current_user.is_admin and logic.settings.get_user_settings(flask_login.current_user.id)['SHOW_INVITATION_LOG']
    if Permissions.GRANT in user_permissions or flask_login.current_user.is_admin:
        project_invitations = logic.projects.get_project_invitations(
            project_id=project_id,
            include_accepted_invitations=show_invitation_log,
            include_expired_invitations=show_invitation_log
        )

    if 'leave' in flask.request.form and Permissions.READ in user_permissions:
        if leave_project_form.validate_on_submit():
            try:
                logic.projects.remove_user_from_project(project_id=project_id, user_id=user_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash('This project does not exist.', 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.UserDoesNotExistError:
                return flask.abort(500)
            except logic.errors.UserNotMemberOfProjectError:
                flask.flash('You have already left the project.', 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                flask.flash('You cannot leave this project, because your are the only user with GRANT permissions.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            else:
                flask.flash('You have successfully left the project.', 'success')
                return flask.redirect(flask.url_for('.projects'))
    if 'delete' in flask.request.form and Permissions.GRANT in user_permissions:
        if delete_project_form.validate_on_submit():
            check_current_user_is_not_readonly()
            try:
                logic.projects.delete_project(project_id=project_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash('This project has already been deleted.', 'success')
                return flask.redirect(flask.url_for('.projects'))
            else:
                flask.flash('You have successfully deleted the project.', 'success')
                return flask.redirect(flask.url_for('.projects'))
    if 'remove_member' in flask.request.form and Permissions.GRANT in user_permissions:
        if remove_project_member_form.validate_on_submit():
            check_current_user_is_not_readonly()
            member_id_str = flask.request.form['remove_member']
            try:
                member_id = int(member_id_str)
            except ValueError:
                flask.flash('The member ID was invalid. Please contact an administrator.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            if member_id == flask_login.current_user.id:
                flask.flash('You cannot remove yourself from a project. Please press "Leave Project" instead.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            try:
                logic.projects.remove_user_from_project(project_id=project_id, user_id=member_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash('This project does not exist.', 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.UserDoesNotExistError:
                flask.flash('This user does not exist.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.UserNotMemberOfProjectError:
                flask.flash('This user is not a member of this project.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                flask.flash('You cannot remove this users from this project, because they are the only user with GRANT permissions.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            else:
                flask.flash('You have successfully removed this user from the project.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'remove_group' in flask.request.form and Permissions.GRANT in user_permissions:
        if remove_project_group_form.validate_on_submit():
            check_current_user_is_not_readonly()
            group_id_str = flask.request.form['remove_group']
            try:
                group_id = int(group_id_str)
            except ValueError:
                flask.flash('The group ID was invalid. Please contact an administrator.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            try:
                logic.projects.remove_group_from_project(project_id=project_id, group_id=group_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash('This project does not exist.', 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.GroupDoesNotExistError:
                flask.flash('This group does not exist.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.GroupNotMemberOfProjectError:
                flask.flash('This group is not a member of this project.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                flask.flash('You cannot remove this group from this project, because they are the only group with GRANT permissions.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            else:
                flask.flash('You have successfully removed this group from the project.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'edit' in flask.request.form and Permissions.WRITE in user_permissions:
        show_edit_form = True
        if edit_project_form.validate_on_submit():
            check_current_user_is_not_readonly()
            try:
                logic.projects.update_project(project_id, edit_project_form.name.data, edit_project_form.description.data)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash('This project does not exist.', 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.ProjectAlreadyExistsError:
                edit_project_form.name.errors.append('A project with this name already exists.')
            except logic.errors.InvalidProjectNameError:
                edit_project_form.name.errors.append('This project name is invalid.')
            else:
                flask.flash('Project information updated successfully.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'add_user' in flask.request.form and Permissions.GRANT in user_permissions:
        if invite_user_form.validate_on_submit():
            check_current_user_is_not_readonly()
            if not any(user.id == invite_user_form.user_id.data for user in invitable_user_list):
                flask.flash('You cannot add this user.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            try:
                other_project_ids = []
                for other_project_id_form in invite_user_form.other_project_ids:
                    try:
                        if other_project_id_form.add_user.data:
                            other_project_ids.append(int(other_project_id_form.project_id.data))
                    except (KeyError, ValueError):
                        pass
                logic.projects.invite_user_to_project(project_id, invite_user_form.user_id.data, flask_login.current_user.id, other_project_ids)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash('This project does not exist.', 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.UserDoesNotExistError:
                flask.flash('This user does not exist.', 'error')
            except logic.errors.UserAlreadyMemberOfProjectError:
                flask.flash('This user is already a member of this project.', 'error')
            else:
                flask.flash('The user was successfully invited to the project.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'add_group' in flask.request.form and Permissions.GRANT in user_permissions:
        if invite_group_form.validate_on_submit():
            check_current_user_is_not_readonly()
            if not any(group.id == invite_group_form.group_id.data for group in invitable_group_list):
                flask.flash('You cannot add this group.', 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            try:
                logic.projects.add_group_to_project(project_id, invite_group_form.group_id.data, permissions=Permissions.READ)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash('This project does not exist.', 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.GroupDoesNotExistError:
                flask.flash('This group does not exist.', 'error')
            except logic.errors.GroupAlreadyMemberOfProjectError:
                flask.flash('This group is already a member of this project.', 'error')
            else:
                flask.flash('The group was successfully added to the project.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'remove_subproject' in flask.request.form and Permissions.GRANT in user_permissions:
        if remove_subproject_form is not None and remove_subproject_form.validate_on_submit():
            check_current_user_is_not_readonly()
            child_project_id = remove_subproject_form.child_project_id.data
            if child_project_id not in child_project_ids:
                flask.flash('Project #{} is not a subproject of this project.'.format(int(child_project_id)), 'error')
            else:
                logic.projects.delete_subproject_relationship(project_id, child_project_id)
                flask.flash('The subproject was successfully removed from this project.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'add_subproject' in flask.request.form and Permissions.GRANT in user_permissions:
        if add_subproject_form is not None and add_subproject_form.validate_on_submit():
            check_current_user_is_not_readonly()
            child_project_id = add_subproject_form.child_project_id.data
            if child_project_id not in addable_project_ids:
                flask.flash('Project #{} cannot become a subproject of this project.'.format(int(child_project_id)), 'error')
            else:
                logic.projects.create_subproject_relationship(project_id, child_project_id, child_can_add_users_to_parent=add_subproject_form.child_can_add_users_to_parent.data)
                flask.flash('The subproject was successfully added to this project.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    return flask.render_template(
        'project.html',
        get_user=logic.users.get_user,
        get_group=logic.groups.get_group,
        get_project=logic.projects.get_project,
        project=project,
        project_member_user_ids=project_member_user_ids,
        project_member_group_ids=project_member_group_ids,
        project_member_user_ids_and_permissions=project_member_user_ids_and_permissions,
        project_member_group_ids_and_permissions=project_member_group_ids_and_permissions,
        project_invitations=project_invitations,
        show_invitation_log=show_invitation_log,
        leave_project_form=leave_project_form,
        delete_project_form=delete_project_form,
        remove_project_member_form=remove_project_member_form,
        remove_project_group_form=remove_project_group_form,
        edit_project_form=edit_project_form,
        show_edit_form=show_edit_form,
        invite_user_form=invite_user_form,
        invitable_user_list=invitable_user_list,
        invite_group_form=invite_group_form,
        invitable_group_list=invitable_group_list,
        show_objects_link=show_objects_link,
        child_project_ids=child_project_ids,
        child_project_ids_can_add_to_parent=child_project_ids_can_add_to_parent,
        parent_project_ids=parent_project_ids,
        add_subproject_form=add_subproject_form,
        addable_projects=addable_projects,
        remove_subproject_form=remove_subproject_form,
        user_may_edit_permissions=Permissions.GRANT in user_permissions
    )


@frontend.route('/projects/', methods=['GET', 'POST'])
@flask_login.login_required
def projects():
    user_id = None
    if 'user_id' in flask.request.args:
        try:
            user_id = int(flask.request.args['user_id'])
        except ValueError:
            pass
    if user_id is not None:
        if user_id != flask_login.current_user.id and not flask_login.current_user.is_admin:
            return flask.abort(403)
        projects = logic.projects.get_user_projects(user_id)
    else:
        projects = logic.projects.get_projects()
    for project in projects:
        project.permissions = logic.projects.get_user_project_permissions(project_id=project.id, user_id=flask_login.current_user.id, include_groups=True)
    create_project_form = CreateProjectForm()
    if create_project_form.name.data is None:
        create_project_form.name.data = ''
    if create_project_form.description.data is None:
        create_project_form.description.data = ''
    show_create_form = False
    if 'create' in flask.request.form:
        show_create_form = True
        if create_project_form.validate_on_submit():
            check_current_user_is_not_readonly()
            try:
                project_id = logic.projects.create_project(create_project_form.name.data, create_project_form.description.data, flask_login.current_user.id).id
            except logic.errors.ProjectAlreadyExistsError:
                create_project_form.name.errors.append('A project with this name already exists.')
            except logic.errors.InvalidProjectNameError:
                create_project_form.name.errors.append('This project name is invalid.')
            else:
                flask.flash('The project has been created successfully.', 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))

    projects_by_id = {
        project.id: project
        for project in projects
    }

    project_id_hierarchy_list = logic.projects.get_project_id_hierarchy_list(list(projects_by_id))

    return flask.render_template(
        "projects.html",
        create_project_form=create_project_form,
        show_create_form=show_create_form,
        Permissions=logic.projects.Permissions,
        projects_by_id=projects_by_id,
        project_id_hierarchy_list=project_id_hierarchy_list
    )


@frontend.route('/projects/<int:project_id>/permissions')
@flask_login.login_required
def project_permissions(project_id):
    try:
        project = logic.projects.get_project(project_id)
    except logic.errors.ProjectDoesNotExistError:
        return flask.abort(404)

    user_permissions = logic.projects.get_project_member_user_ids_and_permissions(project_id, include_groups=False)
    group_permissions = logic.projects.get_project_member_group_ids_and_permissions(project_id)
    if Permissions.GRANT in logic.projects.get_user_project_permissions(project_id=project_id, user_id=flask_login.current_user.id, include_groups=True):
        user_permission_form_data = []
        for user_id, permissions in user_permissions.items():
            if user_id is None:
                continue
            user_permission_form_data.append({'user_id': user_id, 'permissions': permissions.name.lower()})
        group_permission_form_data = []
        for group_id, permissions in group_permissions.items():
            if group_id is None:
                continue
            group_permission_form_data.append({'group_id': group_id, 'permissions': permissions.name.lower()})
        edit_user_permissions_form = ProjectPermissionsForm(user_permissions=user_permission_form_data, group_permissions=group_permission_form_data)
    else:
        edit_user_permissions_form = None
    return flask.render_template('project_permissions.html', project=project, user_permissions=user_permissions, group_permissions=group_permissions, project_permissions=project_permissions, get_user=logic.users.get_user, get_group=logic.groups.get_group, Permissions=Permissions, form=edit_user_permissions_form)


@frontend.route('/projects/<int:project_id>/permissions', methods=['POST'])
@flask_login.login_required
def update_project_permissions(project_id):
    check_current_user_is_not_readonly()
    try:
        if Permissions.GRANT not in logic.projects.get_user_project_permissions(project_id, flask_login.current_user.id, include_groups=True):
            return flask.abort(403)
    except logic.errors.ProjectDoesNotExistError:
        return flask.abort(404)

    edit_user_permissions_form = ProjectPermissionsForm()
    if 'edit_user_permissions' in flask.request.form and edit_user_permissions_form.validate_on_submit():
        # First handle GRANT updates, then others (to prevent temporarily not having a GRANT user)
        for user_permissions_data in sorted(edit_user_permissions_form.user_permissions.data, key=lambda upd: upd['permissions'] != 'grant'):
            user_id = user_permissions_data['user_id']
            try:
                logic.users.get_user(user_id)
            except logic.errors.UserDoesNotExistError:
                continue
            permissions = Permissions.from_name(user_permissions_data['permissions'])
            try:
                logic.projects.update_user_project_permissions(project_id=project_id, user_id=user_id, permissions=permissions)
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                continue
        for group_permissions_data in edit_user_permissions_form.group_permissions.data:
            group_id = group_permissions_data['group_id']
            try:
                logic.groups.get_group(group_id)
            except logic.errors.GroupDoesNotExistError:
                continue
            permissions = Permissions.from_name(group_permissions_data['permissions'])
            logic.projects.update_group_project_permissions(project_id=project_id, group_id=group_id, permissions=permissions)
        flask.flash("Successfully updated project permissions.", 'success')
    else:
        flask.flash("A problem occurred while changing the project permissions. Please try again.", 'error')
    try:
        logic.projects.get_project(project_id)
    except logic.errors.ProjectDoesNotExistError:
        return flask.redirect(flask.url_for('.projects'))
    return flask.redirect(flask.url_for('.project_permissions', project_id=project_id))



# coding: utf-8
"""

"""

from .. import db
from .groups import Group
from .users import User
from .actions import Action
from .permissions import Permissions
from .projects import Project

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class UserActionPermissions(db.Model):
    __tablename__ = 'user_action_permissions'

    action_id = db.Column(db.Integer, db.ForeignKey(Action.id), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(action_id, user_id),
        {},
    )


class GroupActionPermissions(db.Model):
    __tablename__ = 'group_action_permissions'

    action_id = db.Column(db.Integer, db.ForeignKey(Action.id), nullable=False)
    group_id = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(action_id, group_id),
        {},
    )


class ProjectActionPermissions(db.Model):
    __tablename__ = 'project_action_permissions'

    action_id = db.Column(db.Integer, db.ForeignKey(Action.id), nullable=False)
    project_id = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(action_id, project_id),
        {},
    )


class PublicActions(db.Model):
    __tablename__ = 'public_actions'

    action_id = db.Column(db.Integer, db.ForeignKey(Action.id), primary_key=True)



# coding: utf-8
"""

"""

from .. import db


class Tag(db.Model):
    __tablename__ = 'tags'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True, nullable=False)
    uses = db.Column(db.Integer, nullable=False, default=0)



# coding: utf-8
"""

"""

import datetime
import typing

from . import objects
from . import object_permissions
from ..models import ObjectLogEntry, ObjectLogEntryType, Permissions
from .. import db

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def get_object_log_entries(object_id: int, user_id: typing.Optional[int] = None) -> typing.List[ObjectLogEntry]:
    object_log_entries = ObjectLogEntry.query.filter_by(object_id=object_id).order_by(db.desc(ObjectLogEntry.utc_datetime)).all()
    processed_object_log_entries = []
    for object_log_entry in object_log_entries:
        use_object_entry = False
        using_object_type = ''
        if object_log_entry.type == ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT:
            use_object_entry = True
            using_object_type = 'measurement'
        elif object_log_entry.type == ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION:
            use_object_entry = True
            using_object_type = 'sample'
        elif object_log_entry.type == ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA:
            use_object_entry = True
            using_object_type = 'object'
        if use_object_entry:
            using_object_id = using_object_type + '_id'
            object_id = object_log_entry.data[using_object_id]
            object = objects.get_object(object_id=object_id)
            if user_id is not None and Permissions.READ not in object_permissions.get_user_object_permissions(object_id=object_id, user_id=user_id):
                # Clear the using object ID, the user may only know that the
                # object was used for some other object, but not for which
                object_log_entry.data[using_object_id] = None
            else:
                object_log_entry.data[using_object_type] = object

        processed_object_log_entries.append(object_log_entry)
    return processed_object_log_entries


def _store_new_log_entry(type: ObjectLogEntryType, object_id: int, user_id: int, data: dict):
    object_log_entry = ObjectLogEntry(
        type=type,
        object_id=object_id,
        user_id=user_id,
        data=data,
        utc_datetime=datetime.datetime.utcnow()
    )
    db.session.add(object_log_entry)
    db.session.commit()


def create_object(user_id: int, object_id: int, previous_object_id: typing.Optional[int] = None):
    data = {}
    if previous_object_id:
        data['previous_object_id'] = previous_object_id
    _store_new_log_entry(
        type=ObjectLogEntryType.CREATE_OBJECT,
        object_id=object_id,
        user_id=user_id,
        data=data
    )


def edit_object(user_id: int, object_id: int, version_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.EDIT_OBJECT,
        object_id=object_id,
        user_id=user_id,
        data={
            'version_id': version_id
        }
    )


def restore_object_version(user_id: int, object_id: int, version_id: int, restored_version_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.RESTORE_OBJECT_VERSION,
        object_id=object_id,
        user_id=user_id,
        data={
            'version_id': version_id,
            'restored_version_id': restored_version_id
        }
    )


def use_object_in_measurement(user_id: int, object_id: int, measurement_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT,
        object_id=object_id,
        user_id=user_id,
        data={
            'measurement_id': measurement_id
        }
    )


def use_object_in_sample(user_id: int, object_id: int, sample_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION,
        object_id=object_id,
        user_id=user_id,
        data={
            'sample_id': sample_id
        }
    )


def post_comment(user_id: int, object_id: int, comment_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.POST_COMMENT,
        object_id=object_id,
        user_id=user_id,
        data={
            'comment_id': comment_id
        }
    )


def upload_file(user_id: int, object_id: int, file_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.UPLOAD_FILE,
        object_id=object_id,
        user_id=user_id,
        data={
            'file_id': file_id
        }
    )


def create_batch(user_id: int, object_id: int, batch_object_ids: typing.List[int]):
    _store_new_log_entry(
        type=ObjectLogEntryType.CREATE_BATCH,
        object_id=object_id,
        user_id=user_id,
        data={
            'object_ids': batch_object_ids
        }
    )


def assign_location(user_id: int, object_id: int, object_location_assignment_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.ASSIGN_LOCATION,
        object_id=object_id,
        user_id=user_id,
        data={
            'object_location_assignment_id': object_location_assignment_id
        }
    )


def link_publication(
        user_id: int,
        object_id: int,
        doi: str,
        title: typing.Optional[str] = None,
        object_name: typing.Optional[str] = None
):
    _store_new_log_entry(
        type=ObjectLogEntryType.LINK_PUBLICATION,
        object_id=object_id,
        user_id=user_id,
        data={
            'doi': doi,
            'title': title,
            'object_name': object_name
        }
    )


def reference_object_in_metadata(user_id: int, object_id: int, referencing_object_id: int):
    _store_new_log_entry(
        type=ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA,
        object_id=object_id,
        user_id=user_id,
        data={
            'object_id': referencing_object_id
        }
    )



# coding: utf-8
"""
Create default action types.
"""

import os

from ..actions import ActionType

MIGRATION_INDEX = 30
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    default_action_types = [
        {
            'id': ActionType.SAMPLE_CREATION,
            'name': 'Sample Creation',
            'description': 'These Actions represent processes which create a sample.',
            'object_name': 'Sample',
            'object_name_plural': 'Samples',
            'view_text': 'View Samples',
            'perform_text': 'Create Sample',
            'admin_only': False,
            'show_on_frontpage': True,
            'show_in_navbar': True,
            'enable_labels': True,
            'enable_files': True,
            'enable_locations': True,
            'enable_publications': True,
            'enable_comments': True,
            'enable_activity_log': True,
            'enable_related_objects': True
        },
        {
            'id': ActionType.MEASUREMENT,
            'name': 'Measurement',
            'description': 'These Actions represent processes which perform a measurement.',
            'object_name': 'Measurement',
            'object_name_plural': 'Measurements',
            'view_text': 'View Measurements',
            'perform_text': 'Perform Measurement',
            'admin_only': False,
            'show_on_frontpage': True,
            'show_in_navbar': True,
            'enable_labels': False,
            'enable_files': True,
            'enable_locations': True,
            'enable_publications': True,
            'enable_comments': True,
            'enable_activity_log': True,
            'enable_related_objects': True
        },
        {
            'id': ActionType.SIMULATION,
            'name': 'Simulation',
            'description': 'These Actions represent processes which run a simulation.',
            'object_name': 'Simulation',
            'object_name_plural': 'Simulations',
            'view_text': 'View Simulations',
            'perform_text': 'Run Simulation',
            'admin_only': False,
            'show_on_frontpage': True,
            'show_in_navbar': True,
            'enable_labels': False,
            'enable_files': True,
            'enable_locations': True,
            'enable_publications': True,
            'enable_comments': True,
            'enable_activity_log': True,
            'enable_related_objects': True
        }
    ]

    existing_action_type_ids = [
        action_type[0]
        for action_type in db.session.execute("""
            SELECT id
            FROM action_types
        """).fetchall()
    ]

    performed_migration = False
    for action_type in default_action_types:
        # Skip migration by condition
        if action_type['id'] in existing_action_type_ids:
            continue

        # Perform migration
        db.session.execute("""
            INSERT INTO action_types (id, name, description, object_name, object_name_plural, view_text, perform_text, admin_only, show_on_frontpage, show_in_navbar, enable_labels, enable_files, enable_locations, enable_publications, enable_comments, enable_activity_log, enable_related_objects)
            VALUES (:id, :name, :description, :object_name, :object_name_plural, :view_text, :perform_text, :admin_only, :show_on_frontpage, :show_in_navbar, :enable_labels, :enable_files, :enable_locations, :enable_publications, :enable_comments, :enable_activity_log, :enable_related_objects)
        """, params=action_type)
        performed_migration = True
    return performed_migration



# coding: utf-8
"""

"""

import enum
import sqlalchemy.dialects.postgresql as postgresql
from .. import db


@enum.unique
class AuthenticationType(enum.Enum):
    LDAP = 1
    EMAIL = 2
    OTHER = 3
    API_TOKEN = 4


class Authentication(db.Model):
    __tablename__ = "authentications"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'))
    login = db.Column(postgresql.JSONB)
    type = db.Column(db.Enum(AuthenticationType))
    confirmed = db.Column(db.Boolean, default=False, nullable=False)
    user = db.relationship('User', backref="authentication_methods")

    def __init__(self, login, authentication_type, confirmed, user_id):
        self.login = login
        self.type = authentication_type
        self.confirmed = confirmed
        self.user_id = user_id

    def __repr__(self):
        return '<{0}(id={1.id})>'.format(type(self).__name__, self)



# coding: utf-8
"""
Logic module for versioned generic object storage

At its core, this module provides a procedural wrapper of models.objects and
models.versioned_json_object_tables. At the same time, the functions contain
additional logic, e.g. for initial default permissions or object logs.

In practice, the flask-sqlalchemy database engine instance is used. So even
though the underlying models only rely on SQLAlchemy Core and not the ORM,
the functions in this module should be called from within a Flask app context.
"""


import typing
from ..models import Objects, Object, Action, ActionType
from . import object_log, user_log, object_permissions, errors, users, actions, tags
from .notifications import create_notification_for_being_referenced_by_object_metadata
import sqlalchemy.exc


def create_object(
        action_id: int,
        data: dict,
        user_id: int,
        previous_object_id: typing.Optional[int] = None,
        schema: typing.Optional[typing.Dict[str, typing.Any]] = None,
        copy_permissions_object_id: typing.Optional[int] = None
) -> Object:
    """
    Creates an object using the given action and its schema. This function
    also handles logging, object references and default object permissions.

    :param action_id: the ID of an existing action
    :param data: the object's data, which must fit to the action's schema
    :param user_id: the ID of the user who created the object
    :param previous_object_id: the ID of the base object
    :param schema: the object schema used for validation. If schema is None the action schema is used
    :param copy_permissions_object_id: the ID of an existing object to copy
        the permissions from or None
    :return: the created object
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    actions.get_action(action_id)
    users.get_user(user_id)
    try:
        object = Objects.create_object(data=data, schema=schema, user_id=user_id, action_id=action_id)
    except sqlalchemy.exc.IntegrityError:
        raise
    object_log.create_object(object_id=object.object_id, user_id=user_id, previous_object_id=previous_object_id)
    user_log.create_object(object_id=object.object_id, user_id=user_id)
    _update_object_references(object, user_id=user_id)
    _send_user_references_notifications(object, user_id)
    if copy_permissions_object_id is None:
        object_permissions.set_initial_permissions(object)
    else:
        object_permissions.copy_permissions(object.id, copy_permissions_object_id)
        object_permissions.set_user_object_permissions(object.id, user_id, object_permissions.Permissions.GRANT)
    tags.update_object_tag_usage(object)
    return object


def create_object_batch(
        action_id: int,
        data_sequence: typing.Sequence[dict],
        user_id: int,
        copy_permissions_object_id: typing.Optional[int] = None
) -> typing.Sequence[Object]:
    """
    Creates a batch of objects using the given action and its schema. This
    function also handles logging, object references and default object
    permissions. When creating multiple objects for the same action and user
    this function should be used instead of repeatedly calling create_object.

    :param action_id: the ID of an existing action
    :param data_sequence: a sequence containing the objects' data, which must
        fit to the action's schema
    :param user_id: the ID of the user who created the objects
    :param copy_permissions_object_id: the ID of an existing object to copy
        the permissions from or None
    :return: the created objects
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    objects = []
    actions.get_action(action_id)
    users.get_user(user_id)
    try:
        for data in data_sequence:
            try:
                object = Objects.create_object(data=data, schema=None, user_id=user_id, action_id=action_id)
            except sqlalchemy.exc.IntegrityError:
                raise
            objects.append(object)
    finally:
        if objects:
            # objects created before the integrity error must still be accessible
            batch_object_ids = [object.id for object in objects]
            user_log.create_batch(user_id=user_id, batch_object_ids=batch_object_ids)
            for object in objects:
                object_log.create_batch(object_id=object.object_id, user_id=user_id, batch_object_ids=batch_object_ids)
                _update_object_references(object, user_id=user_id)
                _send_user_references_notifications(object, user_id)
                if copy_permissions_object_id is None:
                    object_permissions.set_initial_permissions(object)
                else:
                    object_permissions.copy_permissions(object.id, copy_permissions_object_id)
                    object_permissions.set_user_object_permissions(object.id, user_id, object_permissions.Permissions.GRANT)
                tags.update_object_tag_usage(object)
    return objects


def update_object(object_id: int, data: dict, user_id: int, schema: typing.Optional[typing.Dict[str, typing.Any]] = None) -> None:
    """
    Updates the object to a new version. This function also handles logging
    and object references.

    :param object_id: the ID of the existing object
    :param data: the object's new data, which must fit to the object's schema
    :param user_id: the ID of the user who updated the object
    :param schema: the schema for the new object data
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    object = Objects.update_object(object_id=object_id, data=data, schema=schema, user_id=user_id)
    if object is None:
        raise errors.ObjectDoesNotExistError()
    user_log.edit_object(user_id=user_id, object_id=object.object_id, version_id=object.version_id)
    object_log.edit_object(object_id=object.object_id, user_id=user_id, version_id=object.version_id)
    _update_object_references(object, user_id=user_id)
    _send_user_references_notifications(object, user_id)
    tags.update_object_tag_usage(object)


def restore_object_version(object_id: int, version_id: int, user_id: int) -> None:
    """
    Reverts the changes made to an object and restores a specific version of
    it, while keeping the version history. This function merely adds a new
    version which sets the data and schema to those of the given version.

    :param object_id: the ID of the existing object
    :param version_id: the ID of the object's existing version
    :param user_id: the ID of the user who restored the object version
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.ObjectVersionDoesNotExistError: when an object with the
        given object ID exists, but does not have a version with the given
        version ID
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    object = get_object(object_id=object_id, version_id=version_id)
    object = Objects.update_object(
        object_id=object_id,
        data=object.data,
        schema=object.schema,
        user_id=user_id
    )
    user_log.restore_object_version(user_id=user_id, object_id=object_id, restored_version_id=version_id, version_id=object.version_id)
    object_log.restore_object_version(object_id=object_id, user_id=user_id, restored_version_id=version_id, version_id=object.version_id)
    tags.update_object_tag_usage(object)


def get_object(object_id: int, version_id: typing.Optional[int] = None) -> Object:
    """
    Returns either the current or a specific version of the object.

    :param object_id: the ID of the existing object
    :param version_id: the ID of the object's existing version
    :return: the object with the given object and version IDs
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.ObjectVersionDoesNotExistError: when an object with the
        given object ID exists, but does not have a version with the given
        version ID
    """
    if version_id is None:
        object = Objects.get_current_object(object_id=object_id)
        if object is None:
            raise errors.ObjectDoesNotExistError()
    else:
        object = Objects.get_object_version(object_id=object_id, version_id=version_id)
        if object is None:
            if Objects.get_current_object(object_id=object_id) is None:
                raise errors.ObjectDoesNotExistError()
            else:
                raise errors.ObjectVersionDoesNotExistError()
    return object


def get_object_versions(object_id: int) -> typing.List[Object]:
    """
    Returns all versions of an object, sorted from oldest to newest.

    :param object_id: the ID of the existing object
    :return: the object versions
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    object_versions = Objects.get_object_versions(object_id=object_id)
    if not object_versions:
        raise errors.ObjectDoesNotExistError()
    return object_versions


def get_objects(filter_func=lambda data: True, action_filter=None, **kwargs) -> typing.List[Object]:
    """
    Returns all objects, optionally after filtering the objects by their data
    or by their actions' information.

    :param filter_func: a lambda that may return an SQLAlchemy filter when
        given the object table's data column
    :param action_filter: a SQLAlchemy comparator, used to query only objects
        created by specific actions
    :return: a list of all objects or those matching the given filters
    """
    if action_filter is None:
        action_table = None
    else:
        action_table = Action.__table__
    return Objects.get_current_objects(filter_func=filter_func, action_table=action_table, action_filter=action_filter, **kwargs)


def _get_object_properties(object: Object) -> typing.List[typing.Tuple[typing.List[str], dict, dict]]:
    """
    Returns a list of all properties of an object, as 3-tuples consisting of
    the path to the property, its schema and the actual data.

    :param object: the object
    :return: a list of one 3-tuples for reach of the object's properties
    """
    def iter_object_properties(previous_path, schema, data):
        if schema['type'] == 'object':
            for property_name in schema['properties']:
                if property_name in data:
                    for property in iter_object_properties(previous_path + [property_name], schema['properties'][property_name], data[property_name]):
                        yield property
        elif schema['type'] == 'array':
            for index, item in enumerate(data):
                for property in iter_object_properties(previous_path + [index], schema['items'], item):
                    yield property
        else:
            yield previous_path, schema, data
    return list(iter_object_properties([], object.schema, object.data))


def find_object_references(
        object: Object,
        find_previous_referenced_object_ids: bool = True
) -> typing.List[typing.Tuple[int, typing.Optional[int], str]]:
    """
    Searches for references to other objects.

    :param object: the updated (or newly created) object
    :param find_previous_referenced_object_ids: whether or not to find
        previous referenced object ids
    """
    referenced_object_ids = []
    for path, schema, data in _get_object_properties(object):
        if schema['type'] in ('sample', 'measurement', 'object_reference') and data is not None and data['object_id'] is not None:
            referenced_object_id = data['object_id']
            previous_referenced_object_id = None
            if find_previous_referenced_object_ids and object.version_id > 0:
                previous_object_version = get_object(object.object_id, object.version_id - 1)
                previous_data = previous_object_version.data
                try:
                    for path_element in path:
                        previous_data = previous_data[path_element]
                except (KeyError, IndexError):
                    pass
                else:
                    if previous_data is not None and 'object_id' in previous_data and previous_data['object_id'] is not None:
                        previous_referenced_object_id = previous_data['object_id']
            referenced_object_ids.append((referenced_object_id, previous_referenced_object_id, schema['type']))
    return referenced_object_ids


def _update_object_references(object: Object, user_id: int) -> None:
    """
    Searches for references to other objects and updates these accordingly.

    At this time, only measurements or samples referencing other measurements
    or samples will be handled, adding an entry to the object's log about
    being used in a sample or measurement.

    :param object: the updated (or newly created) object
    :param user_id: the user who caused the object update or creation
    """
    action_type_id = actions.get_action(object.action_id).type_id
    for referenced_object_id, previous_referenced_object_id, schema_type in find_object_references(object):
        if referenced_object_id != previous_referenced_object_id:
            if action_type_id == ActionType.MEASUREMENT and schema_type == 'sample':
                object_log.use_object_in_measurement(object_id=referenced_object_id, user_id=user_id, measurement_id=object.object_id)
            elif action_type_id == ActionType.SAMPLE_CREATION and schema_type == 'sample':
                object_log.use_object_in_sample(object_id=referenced_object_id, user_id=user_id, sample_id=object.object_id)
            else:
                object_log.reference_object_in_metadata(object_id=referenced_object_id, user_id=user_id, referencing_object_id=object.object_id)


def _send_user_references_notifications(object: Object, user_id: int) -> None:
    """
    Searches for references to users and notifies them accordingly.

    :param object: the updated (or newly created) object
    :param user_id: the user who caused the object update or creation
    """

    for referenced_user_id, previous_referenced_user_id in find_user_references(object):
        if referenced_user_id != user_id and referenced_user_id != previous_referenced_user_id:
            create_notification_for_being_referenced_by_object_metadata(referenced_user_id, object.object_id)


def find_user_references(object: Object, find_previous_referenced_user_ids: bool = True) -> typing.List[typing.Tuple[int, typing.Optional[int]]]:
    """
    Searches for references to users.

    :param object: the updated (or newly created) object
    :param find_previous_referenced_user_ids: whether or not to find
        previous referenced user ids
    """
    referenced_user_ids = []
    for path, schema, data in _get_object_properties(object):
        if schema['type'] == 'user' and data is not None and data['user_id'] is not None:
            referenced_user_id = data['user_id']
            previous_referenced_user_id = None
            if find_previous_referenced_user_ids and object.version_id > 0:
                previous_object_version = get_object(object.object_id, object.version_id - 1)
                previous_data = previous_object_version.data
                try:
                    for path_element in path:
                        previous_data = previous_data[path_element]
                except (KeyError, IndexError):
                    pass
                else:
                    if previous_data is not None and 'user_id' in previous_data and previous_data['user_id'] is not None:
                        previous_referenced_user_id = previous_data['user_id']
            referenced_user_ids.append((referenced_user_id, previous_referenced_user_id))
    return referenced_user_ids



# coding: utf-8
"""

"""

from .. import db

association_table = db.Table(
    'user_group_memberships',
    db.metadata,
    db.Column('user_id', db.Integer, db.ForeignKey('users.id')),
    db.Column('group_id', db.Integer, db.ForeignKey('groups.id'))
)


class Group(db.Model):
    __tablename__ = 'groups'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True, nullable=False)
    description = db.Column(db.String, nullable=False, default='')
    members = db.relationship("User", secondary=association_table, backref="groups")


class GroupInvitation(db.Model):
    __tablename__ = 'group_invitations'

    id = db.Column(db.Integer, primary_key=True)
    group_id = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    inviter_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)
    accepted = db.Column(db.Boolean, nullable=False, default=False)



# coding: utf-8
"""
RESTful API for SampleDB
"""

import collections
import base64
import flask

from flask_restful import Resource
import wtforms.validators

from .authentication import object_permissions_required, Permissions
from ...logic.actions import get_action
from ...logic.objects import get_object
from ...logic.files import File, get_file_for_object, get_files_for_object, create_local_file, create_url_file
from ...logic import errors

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def file_info_to_json(file_info: File, include_content: bool = True):
    file_json = {
        'object_id': file_info.object_id,
        'file_id': file_info.id,
        'storage': file_info.storage
    }
    if file_info.storage == 'local':
        file_json.update({
            'original_file_name': file_info.original_file_name
        })
        if include_content:
            with file_info.open() as f:
                base64_content = base64.b64encode(f.read())
            file_json.update({
                'base64_content': base64_content.decode('utf-8')
            })
    if file_info.storage == 'url':
        file_json.update({
            'url': file_info.data['url']
        })
    return file_json


class ObjectFile(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, file_id: int):
        try:
            file_info = get_file_for_object(object_id=object_id, file_id=file_id)
            if file_info is None:
                raise errors.FileDoesNotExistError()
        except errors.FileDoesNotExistError:
            return {
                "message": "file {} of object {} does not exist".format(file_id, object_id)
            }, 404
        if file_info.is_hidden:
            return {
                "message": "file {} of object {} has been hidden".format(file_id, object_id)
            }, 403
        return file_info_to_json(file_info)


class ObjectFiles(Resource):
    @object_permissions_required(Permissions.WRITE)
    def post(self, object_id: int):
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        object = get_object(object_id=object_id)
        action = get_action(action_id=object.action_id)
        if not action.type.enable_files:
            return {
                "message": "Adding files is not enabled for objects of this type"
            }, 403
        if 'object_id' in request_json:
            if request_json['object_id'] != object.object_id:
                return {
                    "message": "object_id must be {}".format(object.object_id)
                }, 400
        if 'storage' not in request_json:
            return {
                "message": "storage must be set"
            }, 400
        storage = request_json['storage']
        if storage not in ('local', 'url'):
            return {
                "message": "storage must be 'local' or 'url'"
            }, 400
        if storage == 'local':
            for key in request_json:
                if key not in {'object_id', 'storage', 'original_file_name', 'base64_content'}:
                    return {
                        "message": "invalid key '{}'".format(key)
                    }, 400
            if 'original_file_name' not in request_json or not request_json['original_file_name']:
                return {
                    "message": "original_file_name must be set for files with local storage"
                }, 400
            original_file_name = request_json['original_file_name']
            if 'base64_content' not in request_json:
                return {
                    "message": "base64_content must be set for files with local storage"
                }, 400
            base64_content = request_json['base64_content']
            try:
                content = base64.b64decode(base64_content.encode('utf-8'), validate=True)
            except Exception:
                return {
                    "message": "base64_content must be base64 encoded"
                }, 400
            file = create_local_file(
                object_id=object_id,
                user_id=flask.g.user.id,
                file_name=original_file_name,
                save_content=lambda stream: stream.write(content)
            )
        if storage == 'url':
            for key in request_json:
                if key not in {'object_id', 'storage', 'url'}:
                    return {
                        "message": "invalid key '{}'".format(key)
                    }, 400
            if 'url' not in request_json:
                return {
                    "message": "url must be set for files with url storage"
                }, 400
            url = request_json['url']
            if not _validate_url(url):
                return {
                    "message": "url must be a valid url"
                }, 400
            file = create_url_file(
                object_id=object_id,
                user_id=flask.g.user.id,
                url=url
            )
        file_url = flask.url_for(
            'api.object_file',
            object_id=file.object_id,
            file_id=file.id
        )
        return flask.redirect(file_url, code=201)

    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int):
        return [
            file_info_to_json(file_info, include_content=False)
            for file_info in get_files_for_object(object_id)
            if not file_info.is_hidden
        ]


def _validate_url(url: str) -> bool:
    """
    Validate a URL using the wtforms.validators.url validator.

    :param url: the URL to validate
    :return: whether or not the URL is valid
    """
    PseudoField = collections.namedtuple('PseudoField', ['data'])
    field = PseudoField(data=url)
    try:
        wtforms.validators.url(message="")(None, field)
    except wtforms.validators.ValidationError:
        return False
    return True



# coding: utf-8
"""

"""

from . import create_instrument
from . import update_instrument
from . import list_instruments
from . import update_instrument_responsible_users
from . import create_action
from . import update_action
from . import list_actions
from . import export_action_schema
from . import create_other_user
from . import send_announcement
from . import set_administrator
from . import set_user_readonly
from . import set_user_hidden
from . import run

script_modules = {
    module.__name__.split('.')[-1]: module
    for module in (
        create_instrument,
        update_instrument,
        list_instruments,
        update_instrument_responsible_users,
        create_action,
        update_action,
        list_actions,
        export_action_schema,
        create_other_user,
        send_announcement,
        set_administrator,
        set_user_readonly,
        set_user_hidden,
        run
    )
}

script_documentation = {
    script_name: script_module.__doc__
    for script_name, script_module
    in script_modules.items()
}

script_functions = {
    script_name: script_module.main
    for script_name, script_module
    in script_modules.items()
}



# coding: utf-8
"""
Script for exporting the schema of an action in SampleDB to a JSON file.

Usage: python -m sampledb export_action_schema <action_id> <schema_file_name>
"""

import json
import sys
from .. import create_app
from ..logic.actions import get_action
from ..logic.errors import ActionDoesNotExistError


def main(arguments):
    if len(arguments) != 2:
        print(__doc__)
        exit(1)
    action_id, schema_file_name = arguments
    try:
        action_id = int(action_id)
    except ValueError:
        print("Error: action_id must be an integer", file=sys.stderr)
        exit(1)

    app = create_app()
    with app.app_context():
        try:
            action = get_action(action_id)
        except ActionDoesNotExistError:
            print('Error: no action with this id exists', file=sys.stderr)
            exit(1)
        schema = action.schema
    with open(schema_file_name, 'w') as schema_file:
        json.dump(schema, schema_file, indent=2)
    print("Success: the action schema has been exported to {}".format(schema_file_name))



# coding: utf-8
"""
Implementation of LDAP authentication.
"""

import ldap3
import ldap3.core.exceptions
import flask

import typing

from ..models import Authentication, AuthenticationType
from .. import db
from . import errors
from . import users
from ..config import LDAP_REQUIRED_CONFIG_KEYS


def is_ldap_configured() -> bool:
    """
    Check whether or not all required LDAP configuration values are set.

    :return: whether all values are set, False otherwise
    """
    return all([
        flask.current_app.config[key]
        for key in LDAP_REQUIRED_CONFIG_KEYS
    ])


def _get_user_dn_and_attributes(user_ldap_uid: str, attributes: typing.Sequence[str] = ()) -> typing.Optional[typing.Sequence[typing.Any]]:
    ldap_host = flask.current_app.config['LDAP_SERVER']
    user_base_dn = flask.current_app.config['LDAP_USER_BASE_DN']
    uid_filter = flask.current_app.config['LDAP_UID_FILTER']
    object_def = flask.current_app.config['LDAP_OBJECT_DEF']
    user_dn = flask.current_app.config['LDAP_USER_DN']
    password = flask.current_app.config['LDAP_PASSWORD']
    server = ldap3.Server(ldap_host, use_ssl=True, get_info=ldap3.ALL)
    try:
        connection = ldap3.Connection(server, user=user_dn, password=password, auto_bind=True)
        object_def = ldap3.ObjectDef(object_def, connection)
        reader = ldap3.Reader(connection, object_def, user_base_dn, uid_filter.format(user_ldap_uid))
        reader.search(attributes)
        # search if uid matches exactly one user, not more
        if len(reader) != 1:
            return None
        user_attributes = [reader[0].entry_dn]
        for attribute in attributes:
            value = getattr(reader[0], attribute, None)
            if value:
                user_attributes.append(value[0])
            else:
                user_attributes.append(None)
        return user_attributes
    except ldap3.core.exceptions.LDAPException:
        return None


def validate_user(user_ldap_uid: str, password: str) -> bool:
    """
    Return whether or not a user with this LDAP uid and password exists.

    This will return False if the uid is not unique, even if the password
    matches one of the users, to avoid conflicts.

    :param user_ldap_uid: the LDAP uid of a user
    :param password: the user's LDAP password
    :return: whether the user credentials are correct or not
    :raise errors.LDAPNotConfiguredError: when LDAP is not configured
    :raise errors.NoEmailInLDAPAccountError: when a user with the UID exists,
        but the LDAP_MAIL_ATTRIBUTE is not set for them
    """
    if not is_ldap_configured():
        raise errors.LDAPNotConfiguredError()

    mail_attribute = flask.current_app.config['LDAP_MAIL_ATTRIBUTE']
    user = _get_user_dn_and_attributes(user_ldap_uid, [mail_attribute])
    if user is None:
        return False
    user_dn, mail = user
    if mail is None:
        raise errors.NoEmailInLDAPAccountError('Email in LDAP-account missing, please contact your administrator')
    ldap_host = flask.current_app.config['LDAP_SERVER']
    # try to bind with user credentials if a matching user exists
    try:
        server = ldap3.Server(ldap_host, use_ssl=True, get_info=ldap3.ALL)
        connection = ldap3.Connection(server, user=user_dn, password=password, raise_exceptions=False)
        return bool(connection.bind())
    except ldap3.core.exceptions.LDAPException:
        return False


def create_user_from_ldap(user_ldap_uid: str) -> typing.Optional[users.User]:
    """
    Create a new user from LDAP information.

    The user's name is read from the LDAP_NAME_ATTRIBUTE, if the attribute
    is set, or the LDAP uid is used otherwise. Their email is read from the
    LDAP_MAIL_ATTRIBUTE.

    :param user_ldap_uid: the LDAP uid of a user
    :return: the newly created user or None, if information is missing
    :raise errors.LDAPNotConfiguredError: when LDAP is not configured
    :raise errors.NoEmailInLDAPAccountError: when the LDAP_MAIL_ATTRIBUTE is
        not set for the user
    """
    if not is_ldap_configured():
        raise errors.LDAPNotConfiguredError()

    name_attribute = flask.current_app.config['LDAP_NAME_ATTRIBUTE']
    mail_attribute = flask.current_app.config['LDAP_MAIL_ATTRIBUTE']
    user = _get_user_dn_and_attributes(
        user_ldap_uid,
        attributes=(name_attribute, mail_attribute)
    )
    if user is None:
        return None
    user_id, name, email = user
    if not email:
        raise errors.NoEmailInLDAPAccountError('There is no email set for your LDAP account. Please contact your administrator.')
    if not name:
        name = user_ldap_uid
    login = user_ldap_uid.lower().strip()
    authentication_methods = Authentication.query.filter(
        db.and_(Authentication.login['login'].astext == login,
                Authentication.type == AuthenticationType.LDAP)
    ).all()
    if not authentication_methods:
        return users.create_user(
            name=name,
            email=email,
            type=users.UserType.PERSON
        )
    return users.get_user(authentication_methods[0].user_id)



# coding: utf-8
"""

"""

import functools
import pint
from ..logic import schemas
from ..logic.units import ureg
from ..logic.errors import ValidationError
from ..logic.schemas.generate_placeholder import generate_placeholder


def form_data_parser(func):
    @functools.wraps(func)
    def wrapper(form_data, schema, id_prefix, errors, required=False):
        try:
            return func(form_data, schema, id_prefix, errors, required=required)
        except ValueError as e:
            for name in form_data:
                if name.startswith(id_prefix) and '__' not in name[len(id_prefix) + 1:]:
                    errors[name] = str(e)
            return None
        except ValidationError as e:
            for name in form_data:
                if name.startswith(id_prefix) and '__' not in name[len(id_prefix) + 1:]:
                    errors[name] = e.message
            return None
    return wrapper


@form_data_parser
def parse_any_form_data(form_data, schema, id_prefix, errors, required=False):
    if schema.get('type') == 'object':
        return parse_object_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'array':
        return parse_array_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'text':
        return parse_text_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'sample':
        return parse_sample_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'measurement':
        return parse_measurement_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'object_reference':
        return parse_object_reference_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'datetime':
        return parse_datetime_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'bool':
        return parse_boolean_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'quantity':
        return parse_quantity_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'tags':
        return parse_tags_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'hazards':
        return parse_hazards_form_data(form_data, schema, id_prefix, errors, required=required)
    elif schema.get('type') == 'user':
        return parse_user_form_data(form_data, schema, id_prefix, errors, required=required)
    raise ValueError('invalid schema')


@form_data_parser
def parse_text_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__text']:
        raise ValueError('invalid text form data')
    text = form_data.get(id_prefix + '__text', [None])[0]
    if text is None and not required:
        return None
    data = {
        '_type': 'text',
        'text': str(text)
    }
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_hazards_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if id_prefix + '__hasnohazards' not in keys:
        raise ValueError('Please select at least one hazard or confirm that the object poses no hazards.')
    hasnohazards = form_data.get(id_prefix + '__hasnohazards', [''])[0]
    if hasnohazards == 'true':
        hasnohazards = True
    elif hasnohazards == 'false':
        hasnohazards = False
    else:
        raise ValueError('invalid hazards form data')
    available_hazards = {
        'ghs01': 1, 'ghs02': 2, 'ghs03': 3, 'ghs04': 4, 'ghs05': 5, 'ghs06': 6, 'ghs07': 7, 'ghs08': 8, 'ghs09': 9
    }
    hazards = []
    for key in keys:
        if key == id_prefix + '__hasnohazards':
            continue
        if key == id_prefix + '__hidden':
            continue
        hazard_name = key.rsplit('__', 1)[1]
        if hazard_name in available_hazards:
            hazards.append(available_hazards[hazard_name])
        else:
            raise ValueError('invalid hazards form data')
    if hasnohazards != (len(hazards) == 0):
        raise ValueError('Please select at least one hazard or confirm that the object poses no hazards.')
    data = {
        '_type': 'hazards',
        'hazards': hazards
    }
    return data


@form_data_parser
def parse_tags_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__tags']:
        raise ValueError('invalid tags form data')
    text = form_data.get(id_prefix + '__tags', [''])[0]
    tags = [tag.lower() for tag in text.split(',')]
    unique_tags = []
    for tag in tags:
        tag = ''.join(c for c in tag if not c.isspace())
        if not tag:
            continue
        if tag not in unique_tags:
            unique_tags.append(tag)
    tags = unique_tags
    data = {
        '_type': 'tags',
        'tags': tags
    }
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_sample_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__oid']:
        raise ValueError('invalid sample form data')
    object_id = form_data.get(id_prefix + '__oid', [''])[0]
    if not object_id:
        if not required:
            return None
        else:
            raise ValueError('Please select a sample.')
    try:
        object_id = int(object_id)
    except ValueError:
        raise ValueError('object_id must be int')
    data = {
        '_type': 'sample',
        'object_id': object_id
    }
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_measurement_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__oid']:
        raise ValueError('invalid measurement form data')
    object_id = form_data.get(id_prefix + '__oid', [''])[0]
    if not object_id:
        if not required:
            return None
        else:
            raise ValueError('Please select a measurement.')
    try:
        object_id = int(object_id)
    except ValueError:
        raise ValueError('object_id must be int')
    data = {
        '_type': 'measurement',
        'object_id': object_id
    }
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_object_reference_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__oid']:
        raise ValueError('invalid object reference form data')
    object_id = form_data.get(id_prefix + '__oid', [''])[0]
    if not object_id:
        if not required:
            return None
        else:
            raise ValueError('Please select an object reference.')
    try:
        object_id = int(object_id)
    except ValueError:
        raise ValueError('object_id must be int')
    data = {
        '_type': 'object_reference',
        'object_id': object_id
    }
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_quantity_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    # TODO: validate schema?
    if set(keys) != {id_prefix + '__magnitude', id_prefix + '__units'} and keys != [id_prefix + '__magnitude']:
        raise ValueError('invalid quantity form data')
    magnitude = form_data[id_prefix + '__magnitude'][0].strip()
    if not magnitude:
        if not required:
            return None
        else:
            raise ValueError('Please enter a magnitude.')
    try:
        magnitude = float(magnitude)
    except ValueError:
        raise ValueError('The magnitude must be a number.')
    if id_prefix + '__units' in form_data:
        units = form_data[id_prefix + '__units'][0]
        try:
            pint_units = ureg.Unit(units)
        except pint.UndefinedUnitError:
            raise ValueError('invalid units')
    else:
        units = '1'
        pint_units = ureg.Unit('1')
    dimensionality = str(pint_units.dimensionality)
    magnitude_in_base_units = ureg.Quantity(magnitude, pint_units).to_base_units().magnitude
    data = {
        '_type': 'quantity',
        'magnitude_in_base_units': magnitude_in_base_units,
        'dimensionality': dimensionality,
        'units': units
    }
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_datetime_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    # TODO: validate schema?
    if keys != [id_prefix + '__datetime']:
        raise ValueError('invalid datetime form data')
    utc_datetime = form_data.get(id_prefix + '__datetime', [''])[0]
    data = {
        '_type': 'datetime',
        'utc_datetime': utc_datetime
    }
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_boolean_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    # TODO: validate schema?
    if set(keys) == {id_prefix + '__hidden', id_prefix + '__value'}:
        value = True
    elif keys == [id_prefix + '__hidden']:
        value = False
    else:
        raise ValueError('invalid boolean form data')
    return {
        '_type': 'bool',
        'value': value
    }


@form_data_parser
def parse_array_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    item_schema = schema['items']
    item_indices = set()
    for key in keys:
        key_rest = key.replace(id_prefix, '', 1)
        if not key_rest.startswith('__'):
            raise ValueError('invalid array form data')
        key_parts = key_rest.split('__')
        if not key_parts[0] == '':
            raise ValueError('invalid array form data')
        try:
            item_index = int(key_parts[1])
        except ValueError:
            raise ValueError('invalid array form data')
        item_indices.add(item_index)
    items = []
    if item_indices:
        num_items = max([i for i in item_indices]) + 1
        for i in range(num_items):
            if i not in item_indices:
                items.append(None)
            else:
                item_id_prefix = id_prefix + '__{}'.format(i)
                items.append(parse_any_form_data(form_data, item_schema, item_id_prefix, errors))
        if None in items:
            # use a placeholder if form_data had no (valid) information on an
            # item, otherwise items would not be a valid array and the form
            # could not be submitted at all
            # if there is no valid placeholder, skip the missing items
            placeholder = generate_placeholder(item_schema)
            if placeholder is None:
                items = [
                    item
                    for item in items
                    if item is not None
                ]
            else:
                items = [
                    placeholder if item is None else item
                    for item in items
                ]
    schemas.validate(items, schema)
    return items


@form_data_parser
def parse_object_form_data(form_data, schema, id_prefix, errors, required=False):
    assert schema['type'] == 'object'
    data = {}
    if any(key.startswith(id_prefix + '__') for key in form_data.keys()):
        for property_name, property_schema in schema['properties'].items():
            property_id_prefix = id_prefix + '__' + property_name
            property_required = (property_name in schema.get('required', []))
            property = parse_any_form_data(form_data, property_schema, property_id_prefix, errors, required=property_required)
            if property is not None:
                data[property_name] = property
    schemas.validate(data, schema)
    return data


@form_data_parser
def parse_user_form_data(form_data, schema, id_prefix, errors, required=False):
    keys = [key for key in form_data.keys() if key.startswith(id_prefix + '__')]
    if keys != [id_prefix + '__uid']:
        raise ValueError('invalid user form data')
    user_id = form_data.get(id_prefix + '__uid', [''])[0]
    if not user_id:
        if not required:
            return None
        else:
            raise ValueError('Please select a user.')
    try:
        user_id = int(user_id)
    except ValueError:
        raise ValueError('user_id must be int')
    data = {
        '_type': 'user',
        'user_id': user_id
    }
    schemas.validate(data, schema)
    return data


def parse_form_data(form_data, schema):
    id_prefix = 'object'
    errors = {}
    data = parse_object_form_data(form_data, schema, id_prefix, errors)
    return data, errors




import base64
import datetime
import io
import json
import typing

import flask
import flask_login
from reportlab.pdfgen.canvas import Canvas
from reportlab.pdfbase.pdfdoc import PDFInfo
from reportlab.lib.pagesizes import A4 as pagesize
from reportlab.lib.units import mm
from PIL import Image
import qrcode
import qrcode.image.pil

from .. import logic
from ..logic import object_log
from ..logic.objects import get_object
from ..logic.actions import get_action
from ..logic.object_log import ObjectLogEntryType
from ..logic.users import get_user

SECTIONS = {
    'activity_log',
    'locations',
    'publications',
    'files',
    'comments'
}

PAGE_WIDTH, PAGE_HEIGHT = pagesize
LEFT_MARGIN = 25.4 * mm
RIGHT_MARGIN = 25.4 * mm
TOP_MARGIN = 31.7 * mm
BOTTOM_MARGIN = 31.7 * mm


def _set_up_page(canvas, object_id, qrcode_uri, logo_uri=None, logo_aspect_ratio=1, logo_alignment='right'):
    canvas.saveState()
    canvas.drawImage(qrcode_uri, PAGE_WIDTH - 27.5 * mm, PAGE_HEIGHT - 27.5 * mm, 20 * mm, 20 * mm)
    if logo_uri:
        if logo_aspect_ratio > 1:
            logo_width = 19 * mm * logo_aspect_ratio
            logo_height = 19 * mm
        else:
            logo_width = 19 * mm
            logo_height = 19 * mm / logo_aspect_ratio
        # right margin due to QR code
        right_margin = 27.5 * mm
        max_width = PAGE_WIDTH - right_margin - LEFT_MARGIN
        if logo_width > max_width:
            logo_height *= max_width / logo_width
            logo_width = max_width
        max_height = 19 * mm
        if logo_height > max_height:
            logo_width *= max_height / logo_height
            logo_height = max_height
        logo_alignment = logo_alignment.lower()
        if logo_alignment == 'left':
            x = LEFT_MARGIN
        elif logo_alignment == 'center':
            x = (LEFT_MARGIN + PAGE_WIDTH - right_margin - logo_width) / 2
        else:
            x = PAGE_WIDTH - right_margin - logo_width
        y = PAGE_HEIGHT - 26.5 * mm
        canvas.drawImage(logo_uri, x, y, logo_width, logo_height)
    canvas.setFont('Helvetica', 8)
    canvas.drawCentredString(PAGE_WIDTH - 17.5 * mm, PAGE_HEIGHT - 29 * mm, "#{}".format(object_id))
    canvas.drawCentredString(PAGE_WIDTH - 17.5 * mm, PAGE_HEIGHT - 33 * mm, "Exported on")
    canvas.drawCentredString(PAGE_WIDTH - 17.5 * mm, PAGE_HEIGHT - 37 * mm, datetime.datetime.utcnow().date().strftime('%Y-%m-%d'))
    page_num = canvas.getPageNumber()
    canvas.setFont('Helvetica', 11)
    canvas.drawCentredString(PAGE_WIDTH / 2, 15 * mm, "{}".format(page_num))
    canvas.restoreState()


def _get_num_fitting_characters(canvas, line, max_width, font_name, font_size):
    line_width = canvas.stringWidth(line, font_name, font_size)
    if line_width <= max_width:
        return len(line)
    # number of characters guaranteed to fit
    lower_boundary = 1
    # maximum possible number of fitting characters
    upper_boundary = len(line) - 1
    approx_num_fitting_characters = len(line)
    while lower_boundary < upper_boundary:
        # approximate fitting characters based on average width
        approx_num_fitting_characters = int(approx_num_fitting_characters / line_width * max_width)
        # enforce upper and lower boundaries for search
        approx_num_fitting_characters = max(lower_boundary + 1, approx_num_fitting_characters)
        approx_num_fitting_characters = min(upper_boundary, approx_num_fitting_characters)
        line_width = canvas.stringWidth(line[:approx_num_fitting_characters], font_name, font_size)
        if line_width < max_width:
            lower_boundary = approx_num_fitting_characters
        elif line_width > max_width:
            upper_boundary = approx_num_fitting_characters - 1
        else:
            lower_boundary = approx_num_fitting_characters
            break
    if ' ' in line[:lower_boundary]:
        return line[:lower_boundary].rfind(' ')
    return lower_boundary


def _draw_left_aligned_wrapped_text(canvas: Canvas, text, left_offset, max_width, top_cursor, font_name, font_size, line_height, justify=False):
    canvas.setFont(font_name, font_size)
    lines = []
    while text:
        text = text.lstrip()
        line = text[:_get_num_fitting_characters(canvas, text, max_width, font_name, font_size)]
        lines.append(line.strip())
        text = text[len(line):]

    for index, line in enumerate(lines):
        if line:
            if justify and len(line) > 1 and index < len(lines) - 1:
                char_space = (max_width - canvas.stringWidth(line, font_name, font_size)) / (len(line) - 1)
            else:
                char_space = 0
            canvas.drawString(left_offset, top_cursor, line, charSpace=char_space)
            top_cursor -= line_height * font_size
            if top_cursor <= BOTTOM_MARGIN:
                canvas.showPage()
                canvas.set_up_page()
                canvas.setFont(font_name, font_size)
                top_cursor = PAGE_HEIGHT - TOP_MARGIN
    return top_cursor


def _append_text(canvas, text, justify=False):
    canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, text, canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica", 11, 1.2, justify)


def _handle_object(data, schema, path, canvas):
    if 'propertyOrder' not in schema:
        for property_name, property_data in data.items():
            property_schema = schema['properties'][property_name]
            _handle_any(property_data, property_schema, path + [property_name], canvas)
    else:
        for property_name in schema['propertyOrder']:
            property_data = data.get(property_name, None)
            property_schema = schema['properties'][property_name]
            _handle_any(property_data, property_schema, path + [property_name], canvas)
        for property_name, property_data in data.items():
            if property_name in schema['propertyOrder']:
                continue
            property_schema = schema['properties'][property_name]
            _handle_any(property_data, property_schema, path + [property_name], canvas)


def _handle_any(data, schema, path, canvas):
    if data is None and schema is None:
        return
    if data is None:
        _handle_none(data, schema, path, canvas)
    elif schema['type'] == 'object':
        _handle_object(data, schema, path, canvas)
    elif schema['type'] == 'array':
        _handle_array(data, schema, path, canvas)
    elif schema['type'] == 'text':
        _handle_text(data, schema, path, canvas)
    elif schema['type'] == 'bool':
        _handle_bool(data, schema, path, canvas)
    elif schema['type'] == 'quantity':
        _handle_quantity(data, schema, path, canvas)
    elif schema['type'] == 'sample':
        _handle_sample(data, schema, path, canvas)
    elif schema['type'] == 'measurement':
        _handle_measurement(data, schema, path, canvas)
    elif schema['type'] == 'object_reference':
        _handle_object_reference(data, schema, path, canvas)
    elif schema['type'] == 'tags':
        _handle_tags(data, schema, path, canvas)
    elif schema['type'] == 'datetime':
        _handle_datetime(data, schema, path, canvas)
    elif schema['type'] == 'hazards':
        _handle_hazards(data, schema, path, canvas)
    elif schema['type'] == 'user':
        _handle_user(data, schema, path, canvas)
    else:
        _handle_unknown(data, schema, path, canvas)


def _handle_none(data, schema, path, canvas):
    text = '• {}: —'.format(schema['title'])
    _append_text(canvas, text)


def _handle_unknown(data, schema, path, canvas):
    text = '• {}: {}'.format(schema['title'], json.dumps(data))
    _append_text(canvas, text)


def _handle_hazards(data, schema, path, canvas):
    if data['hazards']:
        hazards = [
            {
                1: 'Explosive',
                2: 'Flammable',
                3: 'Oxidizing',
                4: 'Compressed Gas',
                5: 'Corrosive',
                6: 'Toxic',
                7: 'Harmful',
                8: 'Health Hazard',
                9: 'Environmental Hazard'
            }.get(hazard, 'Unknown') for hazard in data['hazards']
        ]
        text = '• {}: {}'.format(schema['title'], ', '.join(hazards))
    else:
        text = '• {}: —'.format(schema['title'])
    _append_text(canvas, text)


def _handle_array(data, schema, path, canvas):
    _append_text(canvas, '• {}:'.format(schema['title']))

    previous_left_cursor = canvas.left_cursor
    for index, data in enumerate(data):
        canvas.left_cursor = previous_left_cursor + 2.5 * mm
        _append_text(canvas, '{}. {}:'.format(index + 1, schema['items']['title']))
        canvas.left_cursor = previous_left_cursor + 6 * mm
        _handle_any(data, schema['items'], path + [index], canvas)
    canvas.left_cursor = previous_left_cursor


def _handle_quantity(data, schema, path, canvas):
    if data['units'] == '1':
        text = '• {}: {}'.format(schema['title'], data['magnitude_in_base_units'])
    else:
        quantity = logic.datatypes.Quantity.from_json(data)
        text = '• {}: {:g} {}'.format(schema['title'], quantity.magnitude, logic.units.prettify_units(quantity.units))
    _append_text(canvas, text)


def _handle_sample(data, schema, path, canvas):
    _handle_object_reference(data, schema, path, canvas)


def _handle_measurement(data, schema, path, canvas):
    _handle_object_reference(data, schema, path, canvas)


def _handle_object_reference(data, schema, path, canvas):
    try:
        object = get_object(data['object_id'])
    except logic.errors.ObjectDoesNotExistError:
        object = None
    else:
        object_permissions = logic.object_permissions.get_user_object_permissions(data['object_id'], flask_login.current_user.id)
        if logic.object_permissions.Permissions.READ not in object_permissions:
            object = None
    if object is None:
        text = '• {}: #{}'.format(schema['title'], data['object_id'])
    else:
        text = '• {}: "{}" (#{})'.format(schema['title'], object.data['name']['text'], data['object_id'])
    _append_text(canvas, text)


def _handle_user(data, schema, path, canvas):
    try:
        user = logic.users.get_user(data['user_id'])
        text = '• {}: {} (#{})'.format(schema['title'], user.name, data['user_id'])
    except logic.errors.UserDoesNotExistError:
        text = '• {}: #{}'.format(schema['title'], data['user_id'])
    _append_text(canvas, text)


def _handle_datetime(data, schema, path, canvas):
    text = '• {}: {}'.format(schema['title'], data['utc_datetime'])
    _append_text(canvas, text)


def _handle_tags(data, schema, path, canvas):
    text = '• {}: {}'.format(schema['title'], ', '.join(data['tags']))
    _append_text(canvas, text)


def _handle_text(data, schema, path, canvas):
    prefix = '• {}: '.format(schema['title'])
    text = prefix + data.get('text', '-')
    previously_used_left_cursor = canvas.left_cursor
    indent = canvas.stringWidth(prefix, 'Helvetica', 11)
    for paragraph in text.splitlines(keepends=False):
        if paragraph.strip():
            _append_text(canvas, paragraph, justify=True)
        else:
            canvas.top_cursor -= 1.2 * 11
        canvas.left_cursor = previously_used_left_cursor + indent
    canvas.left_cursor = previously_used_left_cursor


def _handle_bool(data, schema, path, canvas):
    text = '• {}: {}'.format(schema['title'], 'Yes' if data.get('value', True) else 'No')
    _append_text(canvas, text)


def _write_metadata(object, canvas):
    action = get_action(object.action_id)
    object_type = action.type.object_name
    title = '{} #{}: {}'.format(object_type, object.id, object.data.get('name', {}).get('text', ''))

    canvas.bookmarkPage('object_{}'.format(object.id))
    canvas.addOutlineEntry(title, 'object_{}'.format(object.id), level=0)
    canvas.bookmarkPage('object_{}_info'.format(object.id))
    canvas.addOutlineEntry("Information", 'object_{}_info'.format(object.id), level=1)
    canvas.set_up_page()
    canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
    canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, title, canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 18, 1.2)
    canvas.top_cursor -= 6 * mm
    canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, "Information", canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 14, 1.2)
    canvas.top_cursor -= 4 * mm
    _handle_object(object.data, object.schema, [], canvas)
    canvas.showPage()


def _write_activity_log(object, canvas):
    object_log_entries = object_log.get_object_log_entries(object_id=object.id, user_id=flask_login.current_user.id)

    canvas.bookmarkPage('object_{}_activity_log'.format(object.id))
    canvas.addOutlineEntry("Activity Log", 'object_{}_activity_log'.format(object.id), level=1)
    canvas.set_up_page()
    canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
    canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, "Activity Log", canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 14, 1.2)
    canvas.top_cursor -= 4 * mm
    for object_log_entry in reversed(object_log_entries):
        text = '• {} — {}'.format(object_log_entry.utc_datetime.strftime('%Y-%m-%d %H:%M'), get_user(object_log_entry.user_id).name)
        if object_log_entry.type == ObjectLogEntryType.CREATE_BATCH:
            text += ' created this object as part of a batch.'
        elif object_log_entry.type == ObjectLogEntryType.CREATE_OBJECT:
            text += ' created this object.'
        elif object_log_entry.type == ObjectLogEntryType.EDIT_OBJECT:
            text += ' edited this object.'
        elif object_log_entry.type == ObjectLogEntryType.POST_COMMENT:
            text += ' posted a comment.'
        elif object_log_entry.type == ObjectLogEntryType.RESTORE_OBJECT_VERSION:
            text += ' restored a previous version of this object.'
        elif object_log_entry.type == ObjectLogEntryType.UPLOAD_FILE:
            text += ' posted a file.'
        elif object_log_entry.type == ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT:
            if object_log_entry.data['measurement_id'] is not None:
                permissions = logic.object_permissions.get_user_object_permissions(object_log_entry.data['measurement_id'], flask_login.current_user.id)
                if logic.object_permissions.Permissions.READ in permissions:
                    text += ' used this object in measurement "{}" (#{}).'.format(get_object(object_log_entry.data['measurement_id']).data['name']['text'], object_log_entry.data['measurement_id'])
                else:
                    text += ' used this object to create sample #{}.'.format(object_log_entry.data['measurement_id'])
            else:
                text += ' used this object in a measurement.'
        elif object_log_entry.type == ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION:
            if object_log_entry.data['sample_id'] is not None:
                permissions = logic.object_permissions.get_user_object_permissions(object_log_entry.data['sample_id'], flask_login.current_user.id)
                if logic.object_permissions.Permissions.READ in permissions:
                    text += ' used this object to create sample "{}" (#{}).'.format(get_object(object_log_entry.data['sample_id']).data['name']['text'], object_log_entry.data['sample_id'])
                else:
                    text += ' used this object to create sample #{}.'.format(object_log_entry.data['sample_id'])
            else:
                text += ' used this object to create a sample.'
        elif object_log_entry.type == ObjectLogEntryType.ASSIGN_LOCATION:
            object_location_assignment_id = object_log_entry.data['object_location_assignment_id']
            object_location_assignment = logic.locations.get_object_location_assignment(object_location_assignment_id)
            if object_location_assignment.location_id is not None and object_location_assignment.responsible_user_id is not None:
                text += ' assigned this object to location #{} and user #{}.'.format(object_location_assignment.location_id, object_location_assignment.responsible_user_id)
            elif object_location_assignment.location_id is not None:
                text += ' assigned this object to location #{}.'.format(object_location_assignment.location_id)
            elif object_location_assignment.responsible_user_id is not None:
                text += ' assigned this object to user #{}.'.format(object_location_assignment.responsible_user_id)
        elif object_log_entry.type == ObjectLogEntryType.LINK_PUBLICATION:
            text += ' linked publication {} to this object.'.format(object_log_entry.data['doi'])
        elif object_log_entry.type == ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA:
            if object_log_entry.data['object_id'] is not None:
                permissions = logic.object_permissions.get_user_object_permissions(object_log_entry.data['object_id'], flask_login.current_user.id)
                if logic.object_permissions.Permissions.READ in permissions:
                    text += ' referenced this object in the metadata of object "{}" (#{}).'.format(get_object(object_log_entry.data['object_id']).data['name']['text'], object_log_entry.data['object_id'])
                else:
                    text += ' referenced this object in the metadata of object #{}.'.format(object_log_entry.data['object_id'])
            else:
                text += ' referenced this object in the metadata of another object.'
        else:
            text += ' performed an unknown action.'
        _append_text(canvas, text)
    canvas.showPage()


def _write_files(object, canvas):
    files = logic.files.get_files_for_object(object.id)
    if files:
        canvas.bookmarkPage('object_{}_files'.format(object.id))
        canvas.addOutlineEntry("Files", 'object_{}_files'.format(object.id), level=1)
        canvas.set_up_page()
        canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
        canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, "Files", canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 14, 1.2)
        canvas.top_cursor -= 4 * mm
        for file in files:
            if file.is_hidden:
                text = '• {} — {}: (hidden)'.format(file.utc_datetime.strftime('%Y-%m-%d %H:%M'), file.uploader.name)
            else:
                text = '• {} — {}: {}'.format(file.utc_datetime.strftime('%Y-%m-%d %H:%M'), file.uploader.name, file.title)
            _append_text(canvas, text)
        canvas.showPage()
        for index, file in enumerate(files):
            if file.is_hidden:
                continue
            if file.storage != 'local':
                continue
            for file_extension in ('.png', '.jpg', '.jpeg'):
                if file.original_file_name.lower().endswith(file_extension):
                    try:
                        image = Image.open(file.open())
                    except Exception:
                        continue
                    title = "File #{}: {}".format(index + 1, file.title)
                    canvas.bookmarkPage('object_{}_files_{}'.format(object.id, index))
                    canvas.addOutlineEntry(title, 'object_{}_files_{}'.format(object.id, index), level=2)
                    canvas.set_up_page()
                    canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
                    canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, title, canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 12, 1.2)
                    canvas.top_cursor -= 4 * mm
                    if file.description:
                        _append_text(canvas, file.description)
                    original_image_width, original_image_height = image.size
                    image_width = original_image_width / 300 * 25.4 * mm
                    image_height = original_image_height / 300 * 25.4 * mm
                    max_image_width = PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor
                    max_image_height = canvas.top_cursor - BOTTOM_MARGIN
                    if max_image_height / original_image_height <= max_image_width / original_image_width:
                        if max_image_height < image_height:
                            image_width = max_image_height / original_image_height * original_image_width
                            image_height = max_image_height
                    else:
                        if max_image_width < image_width:
                            image_height = max_image_width / original_image_width * original_image_height
                            image_width = max_image_width
                    image_stream = io.BytesIO()
                    image.save(image_stream, format='png')
                    image_stream.seek(0)
                    image_uri = 'data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')
                    canvas.drawImage(image_uri, canvas.left_cursor, canvas.top_cursor - image_height, image_width, image_height)
                    canvas.showPage()
                    break


def _write_comments(object, canvas):
    comments = logic.comments.get_comments_for_object(object.id)
    if comments:
        canvas.bookmarkPage('object_{}_comments'.format(object.id))
        canvas.addOutlineEntry("Comments", 'object_{}_comments'.format(object.id), level=1)
        canvas.set_up_page()
        canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
        canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, "Comments", canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 14, 1.2)
        canvas.top_cursor -= 4 * mm
        for comment in comments:
            text = '• {} — {}:'.format(comment.utc_datetime.strftime('%Y-%m-%d %H:%M'), comment.author.name)
            _append_text(canvas, text)
            previous_left_cursor = canvas.left_cursor
            canvas.left_cursor = previous_left_cursor + 5 * mm
            for paragraph in comment.content.splitlines(keepends=False):
                if paragraph.strip():
                    _append_text(canvas, paragraph, justify=True)
                else:
                    canvas.top_cursor -= 1.2 * 11
            canvas.left_cursor = previous_left_cursor
            canvas.top_cursor -= 4 * mm
        canvas.showPage()


def _write_locations(object, canvas):
    location_assignments = logic.locations.get_object_location_assignments(object.id)
    if location_assignments:
        canvas.bookmarkPage('object_{}_locations'.format(object.id))
        canvas.addOutlineEntry("Locations", 'object_{}_locations'.format(object.id), level=1)
        canvas.set_up_page()
        canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
        canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, "Locations", canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 14, 1.2)
        canvas.top_cursor -= 4 * mm
        for location_assignment in location_assignments:
            text = '• {} — {}:'.format(location_assignment.utc_datetime.strftime('%Y-%m-%d %H:%M'), logic.users.get_user(location_assignment.user_id).name)
            _append_text(canvas, text)
            previous_left_cursor = canvas.left_cursor
            canvas.left_cursor = previous_left_cursor + 5 * mm
            if location_assignment.location_id is not None:
                text = 'Location: {} (#{})'.format(logic.locations.get_location(location_assignment.location_id).name, location_assignment.location_id)
                _append_text(canvas, text)
            if location_assignment.responsible_user_id is not None:
                text = 'Responsible User: {} (#{})'.format(logic.users.get_user(location_assignment.responsible_user_id).name, location_assignment.responsible_user_id)
                _append_text(canvas, text)
            if location_assignment.description:
                _append_text(canvas, 'Description:')
            canvas.left_cursor = previous_left_cursor + 10 * mm
            for paragraph in location_assignment.description.splitlines(keepends=False):
                if paragraph.strip():
                    _append_text(canvas, paragraph, justify=True)
                else:
                    canvas.top_cursor -= 1.2 * 11
            canvas.left_cursor = previous_left_cursor
            canvas.top_cursor -= 4 * mm
        canvas.showPage()


def _write_publications(object, canvas):
    publications = logic.publications.get_publications_for_object(object.id)
    if publications:
        canvas.bookmarkPage('object_{}_publications'.format(object.id))
        canvas.addOutlineEntry("Publications", 'object_{}_publications'.format(object.id), level=1)
        canvas.set_up_page()
        canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
        canvas.top_cursor = _draw_left_aligned_wrapped_text(canvas, "Publications", canvas.left_cursor, PAGE_WIDTH - RIGHT_MARGIN - canvas.left_cursor, canvas.top_cursor, "Helvetica-Bold", 14, 1.2)
        canvas.top_cursor -= 4 * mm
        for publication in publications:
            text = f'• {publication.doi}: {publication.title or "—"}'
            if publication.object_name:
                text += f' ({publication.object_name})'
            _append_text(canvas, text)
        canvas.showPage()


def create_pdfexport(
        object_ids: typing.Sequence[int],
        sections: typing.Set[str] = SECTIONS
) -> bytes:
    """
    Create a PDF containing the exported information of one or more objects.

    :param object_ids: the ID of the objects
    :param sections: a list of sections to include in the generated PDF
    :return: the PDF data
    """
    pdf_stream = io.BytesIO()
    canvas = Canvas(pdf_stream, pagesize=pagesize)
    for object_id in object_ids:
        object = get_object(object_id)

        image = qrcode.make(flask.url_for('.object', object_id=object_id, _external=True), image_factory=qrcode.image.pil.PilImage)
        image_stream = io.BytesIO()
        image.save(image_stream)
        image_stream.seek(0)
        qrcode_uri = 'data:image/svg+xml;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')

        service_name = flask.current_app.config['SERVICE_NAME']
        canvas.setTitle(service_name + ' Export')
        canvas.setAuthor(service_name)
        canvas.setCreator(service_name)
        # set PDF producer (not exposed by reportlab Canvas API)
        PDFInfo.producer = service_name
        canvas.showOutline()
        logo_uri = flask.current_app.config['internal'].get('PDFEXPORT_LOGO_URL', None)
        logo_aspect_ratio = flask.current_app.config['internal'].get('PDFEXPORT_LOGO_ASPECT_RATIO', 1)
        logo_alignment = flask.current_app.config['PDFEXPORT_LOGO_ALIGNMENT']
        canvas.set_up_page = lambda: _set_up_page(canvas, object_id, qrcode_uri, logo_uri, logo_aspect_ratio, logo_alignment)
        canvas.left_cursor = LEFT_MARGIN
        canvas.top_cursor = PAGE_HEIGHT - TOP_MARGIN
        _write_metadata(object, canvas)
        if 'activity_log' in sections:
            _write_activity_log(object, canvas)
        if 'locations' in sections:
            _write_locations(object, canvas)
        if 'publications' in sections:
            _write_publications(object, canvas)
        if 'files' in sections:
            _write_files(object, canvas)
        if 'comments' in sections:
            _write_comments(object, canvas)
    canvas.save()
    pdf_stream.seek(0)
    return pdf_stream.read()



import importlib.util
from setuptools import setup, find_packages
import os.path

setup_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(setup_directory, 'README.md')) as readme_file:
    long_description = readme_file.read()

with open(os.path.join(setup_directory, 'requirements.txt')) as requirements_file:
    requirements = requirements_file.readlines()

version_file_path = os.path.join(setup_directory, 'sampledb', 'version.py')
spec = importlib.util.spec_from_file_location("version", version_file_path)
version = importlib.util.module_from_spec(spec)
spec.loader.exec_module(version)


setup(
    name='sampledb',
    version=version.__version__,
    description='A sample and measurement metadata database',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sciapp/sampledb',
    author='Florian Rhiem',
    author_email='f.rhiem@fz-juelich.de',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Framework :: Flask',
        'Topic :: Scientific/Engineering',
    ],
    packages=find_packages(exclude=['tests', 'tests.*', 'example_data']),
    install_requires=requirements,
    package_data={
        'sampledb': [
            'static/*/*.*',
            'static/*/*/*.*',
            'static/*/*/*/*.*'
        ],
        'sampledb.logic': [
            'unit_definitions.txt'
        ],
        'sampledb.frontend': [
            'templates/*/*.*',
            'templates/*/*/*.*',
            'templates/*/*/*/*.*'
        ]
    }
)



# coding: utf-8
"""
Script for listing all actions in SampleDB.

Usage: python -m sampledb list_actions
"""

from .. import create_app
from ..logic.actions import get_actions


def main(arguments):
    if len(arguments) != 0:
        print(__doc__)
        exit(1)
    app = create_app()
    with app.app_context():
        actions = get_actions()
        for action in actions:
            print(" - #{0.id}: {0.name}".format(action))



# coding: utf-8
"""
Add CREATE_INSTRUMENT_LOG_ENTRY enum value to UserLogEntryType enum.
"""

import os

MIGRATION_INDEX = 18
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    enum_values = db.session.execute("""
        SELECT unnest(enum_range(NULL::userlogentrytype))::text;
    """).fetchall()
    if ('CREATE_INSTRUMENT_LOG_ENTRY',) in enum_values:
        return False

    # Perform migration
    # Use connection and run COMMIT as ALTER TYPE cannot run in a transaction
    connection = db.engine.connect()
    connection.detach()
    connection.execution_options(autocommit=False)
    connection.execute("COMMIT")
    connection.execute("""
        ALTER TYPE userlogentrytype
        ADD VALUE 'CREATE_INSTRUMENT_LOG_ENTRY'
    """)
    connection.close()
    return True



# coding: utf-8
"""
RESTful API for iffSamples
"""

from flask_restful import Api
from .objects import Object, Objects, ObjectVersion, ObjectVersions
from .actions import Action, Actions
from .action_types import ActionType, ActionTypes
from .comments import ObjectComment, ObjectComments
from .files import ObjectFile, ObjectFiles
from .instruments import Instrument, Instruments
from .instrument_log import InstrumentLogEntry, InstrumentLogEntries, InstrumentLogEntryFileAttachment, InstrumentLogEntryFileAttachments, InstrumentLogEntryObjectAttachment, InstrumentLogEntryObjectAttachments, InstrumentLogCategory, InstrumentLogCategories
from .locations import Location, Locations, ObjectLocationAssignment, ObjectLocationAssignments
from .object_permissions import UsersObjectPermissions, UserObjectPermissions, GroupsObjectPermissions, GroupObjectPermissions, ProjectsObjectPermissions, ProjectObjectPermissions, PublicObjectPermissions
from .users import CurrentUser, User, Users

api = Api()
api.add_resource(Objects, '/api/v1/objects/', endpoint='api.objects')
api.add_resource(Object, '/api/v1/objects/<int:object_id>', endpoint='api.object')
api.add_resource(ObjectVersions, '/api/v1/objects/<int:object_id>/versions/', endpoint='api.object_versions')
api.add_resource(ObjectVersion, '/api/v1/objects/<int:object_id>/versions/<int:version_id>', endpoint='api.object_version')
api.add_resource(Actions, '/api/v1/actions/', endpoint='api.actions')
api.add_resource(Action, '/api/v1/actions/<int:action_id>', endpoint='api.action')
api.add_resource(ActionTypes, '/api/v1/action_types/', endpoint='api.action_types')
api.add_resource(ActionType, '/api/v1/action_types/<int(signed=True):type_id>', endpoint='api.action_type')
api.add_resource(ObjectComments, '/api/v1/objects/<int:object_id>/comments/', endpoint='api.object_comments')
api.add_resource(ObjectComment, '/api/v1/objects/<int:object_id>/comments/<int:comment_id>', endpoint='api.object_comment')
api.add_resource(Instruments, '/api/v1/instruments/', endpoint='api.instruments')
api.add_resource(Instrument, '/api/v1/instruments/<int:instrument_id>', endpoint='api.instrument')
api.add_resource(InstrumentLogEntries, '/api/v1/instruments/<int:instrument_id>/log_entries/', endpoint='api.instrument_log_entries')
api.add_resource(InstrumentLogEntry, '/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>', endpoint='api.instrument_log_entry')
api.add_resource(InstrumentLogEntryFileAttachments, '/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/file_attachments/', endpoint='api.instrument_log_entry_file_attachments')
api.add_resource(InstrumentLogEntryFileAttachment, '/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/file_attachments/<int:file_attachment_id>', endpoint='api.instrument_log_entry_file_attachment')
api.add_resource(InstrumentLogEntryObjectAttachments, '/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/object_attachments/', endpoint='api.instrument_log_entry_object_attachments')
api.add_resource(InstrumentLogEntryObjectAttachment, '/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/object_attachments/<int:object_attachment_id>', endpoint='api.instrument_log_entry_object_attachment')
api.add_resource(InstrumentLogCategories, '/api/v1/instruments/<int:instrument_id>/log_categories/', endpoint='api.instrument_log_categories')
api.add_resource(InstrumentLogCategory, '/api/v1/instruments/<int:instrument_id>/log_categories/<int:category_id>', endpoint='api.instrument_log_category')
api.add_resource(Locations, '/api/v1/locations/', endpoint='api.locations')
api.add_resource(Location, '/api/v1/locations/<int:location_id>', endpoint='api.location')
api.add_resource(ObjectFiles, '/api/v1/objects/<int:object_id>/files/', endpoint='api.object_files')
api.add_resource(ObjectFile, '/api/v1/objects/<int:object_id>/files/<int:file_id>', endpoint='api.object_file')
api.add_resource(ObjectLocationAssignments, '/api/v1/objects/<int:object_id>/locations/', endpoint='api.object_location_assignments')
api.add_resource(ObjectLocationAssignment, '/api/v1/objects/<int:object_id>/locations/<int:object_location_assignment_index>', endpoint='api.object_location_assignment')
api.add_resource(UsersObjectPermissions, '/api/v1/objects/<int:object_id>/permissions/users/', endpoint='api.users_object_permissions')
api.add_resource(UserObjectPermissions, '/api/v1/objects/<int:object_id>/permissions/users/<int:user_id>', endpoint='api.user_object_permissions')
api.add_resource(GroupsObjectPermissions, '/api/v1/objects/<int:object_id>/permissions/groups/', endpoint='api.groups_object_permissions')
api.add_resource(GroupObjectPermissions, '/api/v1/objects/<int:object_id>/permissions/groups/<int:group_id>', endpoint='api.group_object_permissions')
api.add_resource(ProjectsObjectPermissions, '/api/v1/objects/<int:object_id>/permissions/projects/', endpoint='api.projects_object_permissions')
api.add_resource(ProjectObjectPermissions, '/api/v1/objects/<int:object_id>/permissions/projects/<int:project_id>', endpoint='api.project_object_permissions')
api.add_resource(PublicObjectPermissions, '/api/v1/objects/<int:object_id>/permissions/public', endpoint='api.public_object_permissions')
api.add_resource(Users, '/api/v1/users/', endpoint='api.users')
api.add_resource(User, '/api/v1/users/<int:user_id>', endpoint='api.user')
api.add_resource(CurrentUser, '/api/v1/users/me', endpoint='api.current_user')



# coding: utf-8
"""
Status route to allow health checks.
"""

import flask


from . import frontend
from .. import db
from .. import version


@frontend.route('/status/')
def status():
    status_info = {
        'sampledb_version': version.__version__,
        'flask_status': True
    }
    try:
        db.engine.execute('select 1').fetchone()
        status_info['postgres_status'] = True
    except Exception:
        status_info['postgres_status'] = False
    if all(status_info.values()):
        status_code = 200
    else:
        status_code = 500
    return flask.jsonify(status_info), status_code
